/*************************************************************************
 *
 * This file is part of the SAMRAI distribution.  For full copyright
 * information, see COPYRIGHT and COPYING.LESSER.
 *
 * Copyright:     (c) 1997-2016 Lawrence Livermore National Security, LLC
 * Description:   Linear time interp operator for node-centered float patch data.
 *
 ************************************************************************/

#ifndef included_pdat_TimeInterpolator
#define included_pdat_TimeInterpolator

#include "TimeInterpolateOperator.h"
#include "SAMRAI/SAMRAI_config.h"
#include "SAMRAI/geom/CartesianGridGeometry.h"

#include <string>

/**
 * Class TimeInterpolator implements standard
 * linear time interpolation for node-centered float patch data.
 * It is derived from the hier::TimeInterpolateOperator base class.
 * The interpolation uses FORTRAN numerical routines.
 *
 * @see hier::TimeInterpolateOperator
 */
using namespace SAMRAI;

class TimeInterpolator:
   public hier::TimeInterpolateOperator
{
public:
   /**
    * Uninteresting default constructor.
    */
   TimeInterpolator(const std::shared_ptr<geom::CartesianGridGeometry >& grid_geom,
                    const std::string discType);

   /**
    * Uninteresting virtual destructor.
    */
   virtual ~TimeInterpolator();

   /**
    * Perform linear time interpolation between two node-centered float
    * patch data sources and place result in the destination patch data.
    * Time interpolation is performed on the intersection of the destination
    * patch data and the input box.  The time to which data is interpolated
    * is provided by the destination data.
    *
    * @pre dynamic_cast<const NodeData<float> *>(&src_data_old) != 0
    * @pre dynamic_cast<const NodeData<float> *>(&src_data_new) != 0
    * @pre dynamic_cast<NodeData<float> *>(&dst_data) != 0
    * @pre (where * src_data_old->getGhostBox()).isSpatiallyEqual(where)
    * @pre (where * src_data_new->getGhostBox()).isSpatiallyEqual(where)
    * @pre (where * dst_data->getGhostBox()).isSpatiallyEqual(where)
    * @pre (dst_data.getDim() == where.getDim()) &&
    *      (dst_data.getDim() == src_data_old.getDim()) &&
    *      (dst_data.getDim() == src_data_new.getDim())
    * @pre direction vector of dst_data is smaller than that of src_data_old or
    *      src_data_new
    */
	void timeInterpolate(
		hier::PatchData& dst_data,
		const hier::Box& where,
		const hier::BoxOverlap& overlap,
		const hier::PatchData& src_data_un,
		const hier::PatchData& src_data_rk1,
		const hier::PatchData& src_data_rk2,
		const hier::PatchData& src_data_rk3,
		const hier::PatchData& src_data_unp1) const;


	void interpolation(
		double simPlat_dt,
		double& un,
		double& rk1,
		double& rk2,
		double& rk3,
		double& unp1) const;


   inline void setStep(int step) {
      this->interp_step = step;
   }

   inline void setRatio(int ratio) {
      this->interp_ratio = ratio;
   }

   inline void setDx(const double* tmpdx) {
      dx = tmpdx[0];
      dy = tmpdx[1];
	dz = tmpdx[2];

   }

  inline void setTimeSubstepNumber(int sn) {
    this->time_substep_number = sn;
  }

private:

  static std::shared_ptr<tbox::Timer> t_interpolate;
  int interp_step;
  int interp_ratio;
  double time_substep_number;

  double xGlower, xGupper, yGlower, yGupper, zGlower, zGupper;

  double dx, dy, dz;

  static double interp_time, current_time, new_time;

  const std::string d_discType;
};

#endif
