#ifndef included_FunctionsXD
#define included_FunctionsXD

#include "SAMRAI/tbox/Utilities.h"
#include <string>
#include <math.h>
#include <vector>
#include "hdf5.h"
#include "Commons.h"


#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))
#define SIGN(X) (((X) > 0) - ((X) < 0))
#define greaterEq(a,b) ((fabs((a) - (b))/1.0E-15 > 10 ? false : (floor(fabs((a) - (b))/1.0E-15) < 1)) || (b)<(a))
#define lessEq(a,b) ((fabs((a) - (b))/1.0E-15 > 10 ? false: (floor(fabs((a) - (b))/1.0E-15) < 1)) || (a)<(b))
#define equalsEq(a,b) ((fabs((a) - (b))/1.0E-15 > 10 ? false: (floor(fabs((a) - (b))/1.0E-15) < 1)))
#define reducePrecision(x, p) (floor(((x) * pow(10, (p)) + 0.5)) / pow(10, (p)))

using namespace external;

/*
 * Calculates coefficients for quintic lagrangian interpolation.
 *    coefs;      coefficients to obtain
 *    coord:      point in which the interpolation must be calculated
 *    position:   surrounding points for interpolation
 */
inline void calculateCoefficientsMapFile(double* coefs, double coord, double* position) {

    for (int i = 0; i < 6; i++) {
        coefs[i] = 0;
    }
    //Search for a perfect fit

    for (int i = 0; i < 6; i++) {
        if (position[i] == coord) {
            coefs[i] = 1;
            return;
        }
    }

    double x1 = position[0];
    double x2 = position[1];
    double x3 = position[2];
    double x4 = position[3];
    double x5 = position[4];
    double x6 = position[5];

    coefs[0] = (coord-x2)*(coord-x3)*(coord-x4)*(coord-x5)*(coord-x6) / ((x1-x2)*(x1-x3)*(x1-x4)*(x1-x5)*(x1-x6));
    coefs[1] = (coord-x1)*(coord-x3)*(coord-x4)*(coord-x5)*(coord-x6) / ((x2-x1)*(x2-x3)*(x2-x4)*(x2-x5)*(x2-x6));
    coefs[2] = (coord-x1)*(coord-x2)*(coord-x4)*(coord-x5)*(coord-x6) / ((x3-x1)*(x3-x2)*(x3-x4)*(x3-x5)*(x3-x6));
    coefs[3] = (coord-x1)*(coord-x2)*(coord-x3)*(coord-x5)*(coord-x6) / ((x4-x1)*(x4-x2)*(x4-x3)*(x4-x5)*(x4-x6));
    coefs[4] = (coord-x1)*(coord-x2)*(coord-x3)*(coord-x4)*(coord-x6) / ((x5-x1)*(x5-x2)*(x5-x3)*(x5-x4)*(x5-x6));
    coefs[5] = (coord-x1)*(coord-x2)*(coord-x3)*(coord-x4)*(coord-x5) / ((x6-x1)*(x6-x2)*(x6-x3)*(x6-x4)*(x6-x5));
}

/*
 * Reads fields from a 3D table interpolating in give coordinates.
 * Uses Trilinear interpolation.
 * Table must be equidistant.
 *    nCoords:            number of coordinates of table
 *    nOutFields:         number of fields to read
 *    coord0, ...:        coordinate values to search in the table
 *    data0, ...:         output variables to store interpolated fields
 *    data0Index, ...:    index of fields inside the table
 *    coord0Data, ...:    variables storing all values of table coordinates
 *    coord0Dx, ...:      spacing of table coordinates
 *    coord0Size, ...:    size of table coordinates
 *    varData, ...:       table data
 *    nVars, ...:         number of variables in the table
 */
inline void readFromFileLinear3D(int nCoords, int nOutFields, double coord0, double coord1, double coord2, double &data0, int data0Index, double *coord0Data, double coord0Dx, int coord0Size, double *coord1Data, double coord1Dx, int coord1Size, double *coord2Data, double coord2Dx, int coord2Size, double *varData, int nVars) {
	// determine spacing parameters of table
	double dx = coord0Dx;
	double dy = coord1Dx;
	double dz = coord2Dx;
	double dxi = 1. / dx;
	double dyi = 1. / dy;
	double dzi = 1. / dz;
	double dxyi  = dxi * dyi;
	double dxzi  = dxi * dzi;
	double dyzi  = dyi * dzi;
	double dxyzi = dxi * dyi * dzi;
	double fh[8];
	double a1, a2, a3, a4, a5, a6, a7, a8;
	// determine location in table
	int idx[8];
	int ix  = 1 + int((coord0 - coord0Data[0] - 1.e-10)/coord0Dx);
	int iy  = 1 + int((coord1 - coord1Data[0] - 1.e-10)/coord1Dx);
	int iz  = 1 + int((coord2 - coord2Data[0] - 1.e-10)/coord2Dx);
	ix = MAX( 1, MIN( ix, coord0Size - 1 ) );
	iy = MAX( 1, MIN( iy, coord1Size - 1 ) );
	iz = MAX( 1, MIN( iz, coord2Size - 1 ) );
	// set up aux vars for interpolation
	double delx = coord0Data[ix] - coord0;
	double dely = coord1Data[iy] - coord1;
	double delz = coord2Data[iz] - coord2;
	idx[0] = data0Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data0Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data0Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data0Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data0Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data0Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data0Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data0Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data0 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
}
/*
 * Reads fields from a 3D table interpolating in give coordinates.
 * Uses Trilinear interpolation.
 * Table must be equidistant.
 *    nCoords:            number of coordinates of table
 *    nOutFields:         number of fields to read
 *    coord0, ...:        coordinate values to search in the table
 *    data0, ...:         output variables to store interpolated fields
 *    data0Index, ...:    index of fields inside the table
 *    coord0Data, ...:    variables storing all values of table coordinates
 *    coord0Dx, ...:      spacing of table coordinates
 *    coord0Size, ...:    size of table coordinates
 *    varData, ...:       table data
 *    nVars, ...:         number of variables in the table
 */
inline void readFromFileLinear3D(int nCoords, int nOutFields, double coord0, double coord1, double coord2, double &data0, double &data1, double &data2, int data0Index, int data1Index, int data2Index, double *coord0Data, double coord0Dx, int coord0Size, double *coord1Data, double coord1Dx, int coord1Size, double *coord2Data, double coord2Dx, int coord2Size, double *varData, int nVars) {
	// determine spacing parameters of table
	double dx = coord0Dx;
	double dy = coord1Dx;
	double dz = coord2Dx;
	double dxi = 1. / dx;
	double dyi = 1. / dy;
	double dzi = 1. / dz;
	double dxyi  = dxi * dyi;
	double dxzi  = dxi * dzi;
	double dyzi  = dyi * dzi;
	double dxyzi = dxi * dyi * dzi;
	double fh[8];
	double a1, a2, a3, a4, a5, a6, a7, a8;
	// determine location in table
	int idx[8];
	int ix  = 1 + int((coord0 - coord0Data[0] - 1.e-10)/coord0Dx);
	int iy  = 1 + int((coord1 - coord1Data[0] - 1.e-10)/coord1Dx);
	int iz  = 1 + int((coord2 - coord2Data[0] - 1.e-10)/coord2Dx);
	ix = MAX( 1, MIN( ix, coord0Size - 1 ) );
	iy = MAX( 1, MIN( iy, coord1Size - 1 ) );
	iz = MAX( 1, MIN( iz, coord2Size - 1 ) );
	// set up aux vars for interpolation
	double delx = coord0Data[ix] - coord0;
	double dely = coord1Data[iy] - coord1;
	double delz = coord2Data[iz] - coord2;
	idx[0] = data0Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data0Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data0Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data0Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data0Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data0Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data0Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data0Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data0 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data1Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data1Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data1Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data1Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data1Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data1Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data1Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data1Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data1 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data2Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data2Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data2Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data2Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data2Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data2Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data2Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data2Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data2 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
}
/*
 * Reads fields from a 3D table interpolating in give coordinates.
 * Uses Trilinear interpolation.
 * Table must be equidistant.
 *    nCoords:            number of coordinates of table
 *    nOutFields:         number of fields to read
 *    coord0, ...:        coordinate values to search in the table
 *    data0, ...:         output variables to store interpolated fields
 *    data0Index, ...:    index of fields inside the table
 *    coord0Data, ...:    variables storing all values of table coordinates
 *    coord0Dx, ...:      spacing of table coordinates
 *    coord0Size, ...:    size of table coordinates
 *    varData, ...:       table data
 *    nVars, ...:         number of variables in the table
 */
inline void readFromFileLinear3D(int nCoords, int nOutFields, double coord0, double coord1, double coord2, double &data0, double &data1, double &data2, double &data3, double &data4, double &data5, double &data6, double &data7, double &data8, int data0Index, int data1Index, int data2Index, int data3Index, int data4Index, int data5Index, int data6Index, int data7Index, int data8Index, double *coord0Data, double coord0Dx, int coord0Size, double *coord1Data, double coord1Dx, int coord1Size, double *coord2Data, double coord2Dx, int coord2Size, double *varData, int nVars) {
	// determine spacing parameters of table
	double dx = coord0Dx;
	double dy = coord1Dx;
	double dz = coord2Dx;
	double dxi = 1. / dx;
	double dyi = 1. / dy;
	double dzi = 1. / dz;
	double dxyi  = dxi * dyi;
	double dxzi  = dxi * dzi;
	double dyzi  = dyi * dzi;
	double dxyzi = dxi * dyi * dzi;
	double fh[8];
	double a1, a2, a3, a4, a5, a6, a7, a8;
	// determine location in table
	int idx[8];
	int ix  = 1 + int((coord0 - coord0Data[0] - 1.e-10)/coord0Dx);
	int iy  = 1 + int((coord1 - coord1Data[0] - 1.e-10)/coord1Dx);
	int iz  = 1 + int((coord2 - coord2Data[0] - 1.e-10)/coord2Dx);
	ix = MAX( 1, MIN( ix, coord0Size - 1 ) );
	iy = MAX( 1, MIN( iy, coord1Size - 1 ) );
	iz = MAX( 1, MIN( iz, coord2Size - 1 ) );
	// set up aux vars for interpolation
	double delx = coord0Data[ix] - coord0;
	double dely = coord1Data[iy] - coord1;
	double delz = coord2Data[iz] - coord2;
	idx[0] = data0Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data0Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data0Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data0Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data0Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data0Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data0Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data0Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data0 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data1Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data1Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data1Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data1Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data1Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data1Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data1Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data1Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data1 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data2Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data2Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data2Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data2Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data2Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data2Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data2Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data2Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data2 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data3Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data3Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data3Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data3Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data3Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data3Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data3Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data3Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data3 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data4Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data4Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data4Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data4Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data4Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data4Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data4Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data4Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data4 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data5Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data5Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data5Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data5Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data5Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data5Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data5Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data5Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data5 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data6Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data6Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data6Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data6Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data6Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data6Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data6Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data6Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data6 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data7Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data7Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data7Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data7Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data7Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data7Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data7Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data7Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data7 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data8Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data8Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data8Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data8Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data8Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data8Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data8Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data8Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	data8 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
}
/*
 * Reads fields from a 3D table interpolating in give coordinates.
 * It also calculates the derivatives of the fields with coordinates.
 * Uses Trilinear interpolation.
 * Table must be equidistant.
 *    nCoords:            number of coordinates of table
 *    nOutFields:         number of fields to read
 *    coord0, ...:        coordinate values to search in the table
 *    data0, ...:         output variables to store interpolated fields
 *    data0Index, ...:    index of fields inside the table
 *    der0Coord0, ...:    output variables where derivatives are calculated
 *    coord0Data, ...:    variables storing all values of table coordinates
 *    coord0Dx, ...:      spacing of table coordinates
 *    coord0Size, ...:    size of table coordinates
 *    varData, ...:       table data
 *    nVars, ...:         number of variables in the table
 */
inline void readFromFileLinearWithDerivatives3D(int nCoords, int nOutFields, double coord0, double coord1, double coord2, double &data0, double &data1, int data0Index, int data1Index, double &der0Coord0, double &der0Coord1, double &der0Coord2, double &der1Coord0, double &der1Coord1, double &der1Coord2, double *coord0Data, double coord0Dx, int coord0Size, double *coord1Data, double coord1Dx, int coord1Size, double *coord2Data, double coord2Dx, int coord2Size, double *varData, int nVars) {
	// determine spacing parameters of table
	double dx = coord0Dx;
	double dy = coord1Dx;
	double dz = coord2Dx;
	double dxi = 1. / dx;
	double dyi = 1. / dy;
	double dzi = 1. / dz;
	double dxyi  = dxi * dyi;
	double dxzi  = dxi * dzi;
	double dyzi  = dyi * dzi;
	double dxyzi = dxi * dyi * dzi;
	double fh[8];
	double a1, a2, a3, a4, a5, a6, a7, a8;
	// determine location in table
	int idx[8];
	int ix  = 1 + int((coord0 - coord0Data[0] - 1.e-10)/coord0Dx);
	int iy  = 1 + int((coord1 - coord1Data[0] - 1.e-10)/coord1Dx);
	int iz  = 1 + int((coord2 - coord2Data[0] - 1.e-10)/coord2Dx);
	ix = MAX( 1, MIN( ix, coord0Size - 1 ) );
	iy = MAX( 1, MIN( iy, coord1Size - 1 ) );
	iz = MAX( 1, MIN( iz, coord2Size - 1 ) );
	// set up aux vars for interpolation
	double delx = coord0Data[ix] - coord0;
	double dely = coord1Data[iy] - coord1;
	double delz = coord2Data[iz] - coord2;
	idx[0] = data0Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data0Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data0Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data0Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data0Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data0Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data0Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data0Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	der0Coord0 = - a2;
	der0Coord1 = - a3;
	der0Coord2 = - a4;
	data0 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data1Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data1Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data1Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data1Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data1Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data1Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data1Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data1Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	der1Coord0 = - a2;
	der1Coord1 = - a3;
	der1Coord2 = - a4;
	data1 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
}
/*
 * Reads fields from a 3D table interpolating in give coordinates.
 * It also calculates the derivatives of the fields with coordinates.
 * Uses Trilinear interpolation.
 * Table must be equidistant.
 *    nCoords:            number of coordinates of table
 *    nOutFields:         number of fields to read
 *    coord0, ...:        coordinate values to search in the table
 *    data0, ...:         output variables to store interpolated fields
 *    data0Index, ...:    index of fields inside the table
 *    der0Coord0, ...:    output variables where derivatives are calculated
 *    coord0Data, ...:    variables storing all values of table coordinates
 *    coord0Dx, ...:      spacing of table coordinates
 *    coord0Size, ...:    size of table coordinates
 *    varData, ...:       table data
 *    nVars, ...:         number of variables in the table
 */
inline void readFromFileLinearWithDerivatives3D(int nCoords, int nOutFields, double coord0, double coord1, double coord2, double &data0, double &data1, double &data2, int data0Index, int data1Index, int data2Index, double &der0Coord0, double &der0Coord1, double &der0Coord2, double &der1Coord0, double &der1Coord1, double &der1Coord2, double &der2Coord0, double &der2Coord1, double &der2Coord2, double *coord0Data, double coord0Dx, int coord0Size, double *coord1Data, double coord1Dx, int coord1Size, double *coord2Data, double coord2Dx, int coord2Size, double *varData, int nVars) {
	// determine spacing parameters of table
	double dx = coord0Dx;
	double dy = coord1Dx;
	double dz = coord2Dx;
	double dxi = 1. / dx;
	double dyi = 1. / dy;
	double dzi = 1. / dz;
	double dxyi  = dxi * dyi;
	double dxzi  = dxi * dzi;
	double dyzi  = dyi * dzi;
	double dxyzi = dxi * dyi * dzi;
	double fh[8];
	double a1, a2, a3, a4, a5, a6, a7, a8;
	// determine location in table
	int idx[8];
	int ix  = 1 + int((coord0 - coord0Data[0] - 1.e-10)/coord0Dx);
	int iy  = 1 + int((coord1 - coord1Data[0] - 1.e-10)/coord1Dx);
	int iz  = 1 + int((coord2 - coord2Data[0] - 1.e-10)/coord2Dx);
	ix = MAX( 1, MIN( ix, coord0Size - 1 ) );
	iy = MAX( 1, MIN( iy, coord1Size - 1 ) );
	iz = MAX( 1, MIN( iz, coord2Size - 1 ) );
	// set up aux vars for interpolation
	double delx = coord0Data[ix] - coord0;
	double dely = coord1Data[iy] - coord1;
	double delz = coord2Data[iz] - coord2;
	idx[0] = data0Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data0Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data0Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data0Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data0Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data0Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data0Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data0Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	der0Coord0 = - a2;
	der0Coord1 = - a3;
	der0Coord2 = - a4;
	data0 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data1Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data1Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data1Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data1Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data1Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data1Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data1Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data1Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	der1Coord0 = - a2;
	der1Coord1 = - a3;
	der1Coord2 = - a4;
	data1 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
	idx[0] = data2Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz)));
	idx[1] = data2Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz)));
	idx[2] = data2Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[3] = data2Index + nVars * (ix + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[4] = data2Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz)));
	idx[5] = data2Index + nVars * (ix - 1 + coord0Size * (iy + coord1Size * (iz - 1)));
	idx[6] = data2Index + nVars * (ix + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	idx[7] = data2Index + nVars * (ix - 1 + coord0Size * (iy - 1 + coord1Size * (iz - 1)));
	// set up aux vars for interpolation
	fh[0] = varData[idx[0]];
	fh[1] = varData[idx[1]];
	fh[2] = varData[idx[2]];
	fh[3] = varData[idx[3]];
	fh[4] = varData[idx[4]];
	fh[5] = varData[idx[5]];
	fh[6] = varData[idx[6]];
	fh[7] = varData[idx[7]];
	// set up coeffs of interpolation polynomical and evaluate function values
	a1 = fh[0];
	a2 = dxi   * ( fh[1] - fh[0] );
	a3 = dyi   * ( fh[2] - fh[0] );
	a4 = dzi   * ( fh[3] - fh[0] );
	a5 = dxyi  * ( fh[4] - fh[1] - fh[2] + fh[0] );
	a6 = dxzi  * ( fh[5] - fh[1] - fh[3] + fh[0] );
	a7 = dyzi  * ( fh[6] - fh[2] - fh[3] + fh[0] );
	a8 = dxyzi * ( fh[7] - fh[0] + fh[1] + fh[2] + fh[3] - fh[4] - fh[5] - fh[6] );
	der2Coord0 = - a2;
	der2Coord1 = - a3;
	der2Coord2 = - a4;
	data2 = a1 +  a2 * delx +  a3 * dely +  a4 * delz +  a5 * delx * dely +  a6 * delx * delz +  a7 * dely * delz +  a8 * delx * dely * delz;
}

/*
 * Given a field in the table and all coordinates but one, it finds the remnaining coordinate that fits better with the values.
 *    nCoords:            number of coordinates of table
 *    nInFields:          number of fields given
 *    coord0, ...:        coordinate values to search in the table
 *    datain:             field used to search the coordinate
 *    dataIndex:          index of field in the table
 *    coord0Data, ...:    variables storing all values of table coordinates
 *    coord0Dx, ...:      spacing of table coordinates
 *    coord0Size, ...:    size of table coordinates
 *    varData, ...:       table data
 *    nVars, ...:         number of variables in the table
 */
inline double findCoordFromFile3D_1(int nCoords, int nInFields, double coord0, double coord1, double coord2, double datain, int dataIndex, double* coord0Data, double coord0Dx, int coord0Size, double* coord1Data, double coord1Dx, int coord1Size, double* coord2Data, double coord2Dx, int coord2Size, double *varData, int nVars) {
	// local vars
	double cmax = coord1Data[coord1Size-1]; // max 
	double cmin = coord1Data[0]; // min

	// setting up some vars
	double data0 = datain;
	double data1 = data0;
	double data, data2;
	double c0 = coord1; 
	double c = coord1; 
	double c1 = c; 

	// step 1: do we already have the right temperature
	readFromFileLinear3D(nCoords, nInFields, coord0, c, coord2, data, dataIndex, coord0Data, coord0Dx, coord0Size, coord1Data, coord1Dx, coord1Size, coord2Data, coord2Dx, coord2Size, varData, nVars);
	if(fabs(data-data0) < 1e-6*fabs(data0)) {
		return coord1;
	}
	// try bisection
	int itmax = 100;
	// temporary local vars
	double c2;
	double data1a;
	double data2a;
	// prepare
	c = c0; 
	// Bisection by value
	c1 = MIN(c0 + 2 * coord1Dx, cmax);
	c2 = MAX(c0 - 2 * coord1Dx, cmin);
	readFromFileLinear3D(nCoords, nInFields, coord0, c1, coord2, data1a, dataIndex, coord0Data, coord0Dx, coord0Size, coord1Data, coord1Dx, coord1Size, coord2Data, coord2Dx, coord2Size, varData, nVars);
	readFromFileLinear3D(nCoords, nInFields, coord0, c2, coord2, data2a, dataIndex, coord0Data, coord0Dx, coord0Size, coord1Data, coord1Dx, coord1Size, coord2Data, coord2Dx, coord2Size, varData, nVars);
	data1=data1a-data0;
	data2=data2a-data0;
	int bcount = 0;
	int maxbcount = 80;
	bool notBracketed = false;
	while(data1*data2 >= 0.0) {
		c1 = MIN(c1 + 2 * coord1Dx,cmax);
		c2 = MAX(c2 - 2 * coord1Dx,cmin);
		readFromFileLinear3D(nCoords, nInFields, coord0, c1, coord2, data1a, dataIndex, coord0Data, coord0Dx, coord0Size, coord1Data, coord1Dx, coord1Size, coord2Data, coord2Dx, coord2Size, varData, nVars);
		readFromFileLinear3D(nCoords, nInFields, coord0, c2, coord2, data2a, dataIndex, coord0Data, coord0Dx, coord0Size, coord1Data, coord1Dx, coord1Size, coord2Data, coord2Dx, coord2Size, varData, nVars);
		data1=data1a-data0;
		data2=data2a-data0;
		bcount++;
		if(bcount >= maxbcount) {
			notBracketed = true;
			break;
		}
	} 
	if (!notBracketed) {
		//data1 and data2 may have been modified in loop above, so we check here to see if either is zero.
		if (equalsEq(data1, 0.0)) {
			return c1;
		}
		if (equalsEq(data2, 0.0)) {
			return c2;
		}
		double dc, cmid, fmid;
		if(data1 < 0.0) {
			c = c1;
			dc = c2 - c1;
		} else {
			c = c2;
			dc = c1 - c2;
		}
		for(int it=0;it<itmax;it++) {
			dc = dc * 0.5;
			cmid = c + dc;
			readFromFileLinear3D(nCoords, nInFields, coord0, cmid, coord2, data, dataIndex, coord0Data, coord0Dx, coord0Size, coord1Data, coord1Dx, coord1Size, coord2Data, coord2Dx, coord2Size, varData, nVars);
			fmid=data-data0;
			if(fmid <= 0.0) c=cmid;
			if(fabs(data - data0) <= fabs(1e-10*data0)) {
				return cmid;
			}
		} 
	}
	//Try bisection by index
	int i0  = int((c0 - coord1Data[0]) / coord1Dx);  // index of point closest to lto
	int ilo = MIN(MAX(0, i0), coord1Size-2);
	int ihi = MAX(MIN(coord1Size-1, i0 + 1), 1);

	c1 = coord1Data[ilo];
	c2 = coord1Data[ihi];
	readFromFileLinear3D(nCoords, nInFields, coord0, c1, coord2, data1a, dataIndex, coord0Data, coord0Dx, coord0Size, coord1Data, coord1Dx, coord1Size, coord2Data, coord2Dx, coord2Size, varData, nVars);
	readFromFileLinear3D(nCoords, nInFields, coord0, c2, coord2, data2a, dataIndex, coord0Data, coord0Dx, coord0Size, coord1Data, coord1Dx, coord1Size, coord2Data, coord2Dx, coord2Size, varData, nVars);
	data1=data1a-data0;
	data2=data2a-data0;
	// iterate until we bracket the right eps, but enforce
	// ddata/dc > 0, so eps(c1) > eps(c2)
	if(data1*data2 >= 0.0) {
		if (data1 > 0.0) {
			int k = 0;
			while (ilo >= 1 && data1 > 0.0) {
				ilo = MAX(0,ilo - 2*k - 1);
				c1 = coord1Data[ilo];
				readFromFileLinear3D(nCoords, nInFields, coord0, c1, coord2, data1a, dataIndex, coord0Data, coord0Dx, coord0Size, coord1Data, coord1Dx, coord1Size, coord2Data, coord2Dx, coord2Size, varData, nVars);
				data1 = data1a - data0;
				k++;
			}
			if (data1 > 0.0) {
				return coord1Data[0];
			}
		}
		if (data2 < 0.0) {
			int k = 0;
			while (ihi <= coord1Size-2 && data2 < 0.0) {
				ihi = MIN(coord1Size - 1, ihi + 2*k + 1);
				c2 = coord1Data[ihi];
				readFromFileLinear3D(nCoords, nInFields, coord0, c2, coord2, data2a, dataIndex, coord0Data, coord0Dx, coord0Size, coord1Data, coord1Dx, coord1Size, coord2Data, coord2Dx, coord2Size, varData, nVars);
				data2 = data2a - data0;
				k++;
			}
			if (data2 < 0.0) {
				return coord1Data[coord1Size - 1];
			}
		}
	}
	//data1 and data2 may have been modified in loop above, so we check here to see if either is zero.
	if (equalsEq(data1, 0.0)) {
		return coord1Data[ilo];
	}
	if (equalsEq(data2, 0.0)) {
		return coord1Data[ihi];
	}

	int jj = 0;
	bool found = false;
	if ((ihi - ilo) == 1) {
		found = true;
	}

	itmax = 10;
	while (jj < itmax && !found) {
		int ii = (ihi + ilo) / 2;
		double ctmp = coord1Data[ii];
		readFromFileLinear3D(nCoords, nInFields, coord0, ctmp, coord2, data, dataIndex, coord0Data, coord0Dx, coord0Size, coord1Data, coord1Dx, coord1Size, coord2Data, coord2Dx, coord2Size, varData, nVars);
		if (fabs(data - data0) < 1.0e-15*fabs(data0)) {
			return ctmp;
		} else if (data < data0) {
			ilo = ii;
			data1a = data;
		} else if (data > data0) {
			ihi = ii;
			data2a = data;
		}
		if ((ihi - ilo) == 1) {
			found = true;
		}
		jj++;
	}

	if (found) {
		//The root has been bracketed.  Now use linear interpolation to find the value.
		return (data0 - data1a)*(coord1Data[ihi] - coord1Data[ilo]) / (data2a - data1a) + coord1Data[ilo];
	}

	TBOX_ERROR("findCoordFromFile3D_1: cannot find the coordinate "<<coord1<<" from "<<data);
}

#define vector(v, i, j, k) (v)[i+ilast*(j+jlast*(k))]

#define POLINT_MACRO_LINEAR_1(y1, y2, i_ext, s_ext, dx, simPlat_dt, ilast, jlast) (-i_ext*y1 + i_ext*y2 + s_ext*y2)
#define POLINT_MACRO_LINEAR_2(y1, y2, i_ext, s_ext, dx, simPlat_dt, ilast, jlast) (-i_ext*y1 + i_ext*y2 - s_ext*y1)
#define POLINT_MACRO_QUADRATIC_1(y1,y2,y3, i_ext, s_ext, dx, simPlat_dt, ilast, jlast) (0.5*(i_ext*i_ext*y1-2*i_ext*i_ext*y2+i_ext*i_ext*y3+i_ext*s_ext*y1-4*i_ext*s_ext*y2+3*i_ext*s_ext*y3+2*s_ext*s_ext*y3)/(s_ext*s_ext))
#define POLINT_MACRO_QUADRATIC_2(y1,y2,y3, i_ext, s_ext, dx, simPlat_dt, ilast, jlast) (0.5*(i_ext*i_ext*y1-2*i_ext*i_ext*y2+i_ext*i_ext*y3+3*i_ext*s_ext*y1-4*i_ext*s_ext*y2+i_ext*s_ext*y3+2*s_ext*s_ext*y1)/(s_ext*s_ext))
#define POLINT_MACRO_CUBIC_1(y1,y2,y3, y4, i_ext, s_ext, dx, simPlat_dt, ilast, jlast) (-(1.0/6.0)*(i_ext*i_ext*i_ext*y1-3*i_ext*i_ext*i_ext*y2+3*i_ext*i_ext*i_ext*y3-i_ext*i_ext*i_ext*y4+3*i_ext*i_ext*s_ext*y1-12*i_ext*i_ext*s_ext*y2+15*i_ext*i_ext*s_ext*y3-6*i_ext*i_ext*s_ext*y4+2*i_ext*s_ext*s_ext*y1-9*i_ext*s_ext*s_ext*y2+18*i_ext*s_ext*s_ext*y3-11*i_ext*s_ext*s_ext*y4-6*s_ext*s_ext*s_ext*y4)/(s_ext*s_ext*s_ext))
#define POLINT_MACRO_CUBIC_2(y1,y2,y3, y4, i_ext, s_ext, dx, simPlat_dt, ilast, jlast) ((1.0/6.0)*(i_ext*i_ext*i_ext*y1-3*i_ext*i_ext*i_ext*y2+3*i_ext*i_ext*i_ext*y3-i_ext*i_ext*i_ext*y4+6*i_ext*i_ext*s_ext*y1-15*i_ext*i_ext*s_ext*y2+12*i_ext*i_ext*s_ext*y3-3*i_ext*i_ext*s_ext*y4+11*i_ext*s_ext*s_ext*y1-18*i_ext*s_ext*s_ext*y2+9*i_ext*s_ext*s_ext*y3-2*i_ext*s_ext*s_ext*y4+6*s_ext*s_ext*s_ext*y1)/(s_ext*s_ext*s_ext))

#define D1CDO4_i(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) ((((-vector(paru, (pari) + 2, (parj), (park))) + 8.0 * vector(paru, (pari) + 1, (parj), (park))) + ((-8.0 * vector(paru, (pari) - 1, (parj), (park))) + vector(paru, (pari) - 2, (parj), (park)))) / (12.0 * dx[0]))

#define D1CDO4_j(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) ((((-vector(paru, (pari), (parj) + 2, (park))) + 8.0 * vector(paru, (pari), (parj) + 1, (park))) + ((-8.0 * vector(paru, (pari), (parj) - 1, (park))) + vector(paru, (pari), (parj) - 2, (park)))) / (12.0 * dx[1]))

#define D1CDO4_k(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) ((((-vector(paru, (pari), (parj), (park) + 2)) + 8.0 * vector(paru, (pari), (parj), (park) + 1)) + ((-8.0 * vector(paru, (pari), (parj), (park) - 1)) + vector(paru, (pari), (parj), (park) - 2))) / (12.0 * dx[2]))

#define lieforward_i(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (((-3.0 * vector(paru, (pari) - 1, (parj), (park))) + (-10.0 * vector(paru, (pari), (parj), (park))) + 18.0 * vector(paru, (pari) + 1, (parj), (park)) + (-6.0 * vector(paru, (pari) + 2, (parj), (park))) + vector(paru, (pari) + 3, (parj), (park))) / (dx[0] * 12.0))

#define lieforward_j(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (((-3.0 * vector(paru, (pari), (parj) - 1, (park))) + (-10.0 * vector(paru, (pari), (parj), (park))) + 18.0 * vector(paru, (pari), (parj) + 1, (park)) + (-6.0 * vector(paru, (pari), (parj) + 2, (park))) + vector(paru, (pari), (parj) + 3, (park))) / (dx[1] * 12.0))

#define lieforward_k(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (((-3.0 * vector(paru, (pari), (parj), (park) - 1)) + (-10.0 * vector(paru, (pari), (parj), (park))) + 18.0 * vector(paru, (pari), (parj), (park) + 1) + (-6.0 * vector(paru, (pari), (parj), (park) + 2)) + vector(paru, (pari), (parj), (park) + 3)) / (dx[2] * 12.0))

#define liebackward_i(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) ((((-vector(paru, (pari) - 3, (parj), (park))) + 6.0 * vector(paru, (pari) - 2, (parj), (park))) - 18.0 * vector(paru, (pari) - 1, (parj), (park)) + (10.0 * vector(paru, (pari), (parj), (park)) + 3.0 * vector(paru, (pari) + 1, (parj), (park)))) / (dx[0] * 12.0))

#define liebackward_j(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) ((((-vector(paru, (pari), (parj) - 3, (park))) + 6.0 * vector(paru, (pari), (parj) - 2, (park))) - 18.0 * vector(paru, (pari), (parj) - 1, (park)) + (10.0 * vector(paru, (pari), (parj), (park)) + 3.0 * vector(paru, (pari), (parj) + 1, (park)))) / (dx[1] * 12.0))

#define liebackward_k(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) ((((-vector(paru, (pari), (parj), (park) - 3)) + 6.0 * vector(paru, (pari), (parj), (park) - 2)) - 18.0 * vector(paru, (pari), (parj), (park) - 1) + (10.0 * vector(paru, (pari), (parj), (park)) + 3.0 * vector(paru, (pari), (parj), (park) + 1))) / (dx[2] * 12.0))

#define D2CDO4_i(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) ((((-vector(paru, (pari) + 2, (parj), (park))) + 16.0 * vector(paru, (pari) + 1, (parj), (park))) - 30.0 * vector(paru, (pari), (parj), (park)) + 16.0 * vector(paru, (pari) - 1, (parj), (park)) - vector(paru, (pari) - 2, (parj), (park))) / (12.0 * (dx[0] * dx[0])))

#define D2CDO4_j(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) ((((-vector(paru, (pari), (parj) + 2, (park))) + 16.0 * vector(paru, (pari), (parj) + 1, (park))) - 30.0 * vector(paru, (pari), (parj), (park)) + 16.0 * vector(paru, (pari), (parj) - 1, (park)) - vector(paru, (pari), (parj) - 2, (park))) / (12.0 * (dx[1] * dx[1])))

#define D2CDO4_k(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) ((((-vector(paru, (pari), (parj), (park) + 2)) + 16.0 * vector(paru, (pari), (parj), (park) + 1)) - 30.0 * vector(paru, (pari), (parj), (park)) + 16.0 * vector(paru, (pari), (parj), (park) - 1) - vector(paru, (pari), (parj), (park) - 2)) / (12.0 * (dx[2] * dx[2])))

#define D1CDO4crossed_ii(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (((vector(paru, (pari) - 2, (parj), (park)) + 8.0 * vector(paru, (pari) + 1, (parj), (park)) + 64.0 * vector(paru, (pari) - 1, (parj), (park)) + 8.0 * vector(paru, (pari) + 2, (parj), (park)) + 8.0 * vector(paru, (pari) - 2, (parj), (park)) + 64.0 * vector(paru, (pari) + 1, (parj), (park)) + 8.0 * vector(paru, (pari) - 1, (parj), (park)) + vector(paru, (pari) + 2, (parj), (park))) - (8.0 * vector(paru, (pari) - 1, (parj), (park)) + vector(paru, (pari) + 2, (parj), (park)) + 8.0 * vector(paru, (pari) - 2, (parj), (park)) + 64.0 * vector(paru, (pari) + 1, (parj), (park)) + 64.0 * vector(paru, (pari) - 1, (parj), (park)) + 8.0 * vector(paru, (pari) + 2, (parj), (park)) + vector(paru, (pari) - 2, (parj), (park)) + 8.0 * vector(paru, (pari) + 1, (parj), (park)))) / (144.0 * dx[0] * dx[0]))

#define D1CDO4crossed_ij(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (((vector(paru, (pari) - 2, (parj) - 2, (park)) + 8.0 * vector(paru, (pari) + 1, (parj) - 2, (park)) + 64.0 * vector(paru, (pari) - 1, (parj) - 1, (park)) + 8.0 * vector(paru, (pari) + 2, (parj) - 1, (park)) + 8.0 * vector(paru, (pari) - 2, (parj) + 1, (park)) + 64.0 * vector(paru, (pari) + 1, (parj) + 1, (park)) + 8.0 * vector(paru, (pari) - 1, (parj) + 2, (park)) + vector(paru, (pari) + 2, (parj) + 2, (park))) - (8.0 * vector(paru, (pari) - 1, (parj) - 2, (park)) + vector(paru, (pari) + 2, (parj) - 2, (park)) + 8.0 * vector(paru, (pari) - 2, (parj) - 1, (park)) + 64.0 * vector(paru, (pari) + 1, (parj) - 1, (park)) + 64.0 * vector(paru, (pari) - 1, (parj) + 1, (park)) + 8.0 * vector(paru, (pari) + 2, (parj) + 1, (park)) + vector(paru, (pari) - 2, (parj) + 2, (park)) + 8.0 * vector(paru, (pari) + 1, (parj) + 2, (park)))) / (144.0 * dx[0] * dx[1]))

#define D1CDO4crossed_ik(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (((vector(paru, (pari) - 2, (parj), (park) - 2) + 8.0 * vector(paru, (pari) + 1, (parj), (park) - 2) + 64.0 * vector(paru, (pari) - 1, (parj), (park) - 1) + 8.0 * vector(paru, (pari) + 2, (parj), (park) - 1) + 8.0 * vector(paru, (pari) - 2, (parj), (park) + 1) + 64.0 * vector(paru, (pari) + 1, (parj), (park) + 1) + 8.0 * vector(paru, (pari) - 1, (parj), (park) + 2) + vector(paru, (pari) + 2, (parj), (park) + 2)) - (8.0 * vector(paru, (pari) - 1, (parj), (park) - 2) + vector(paru, (pari) + 2, (parj), (park) - 2) + 8.0 * vector(paru, (pari) - 2, (parj), (park) - 1) + 64.0 * vector(paru, (pari) + 1, (parj), (park) - 1) + 64.0 * vector(paru, (pari) - 1, (parj), (park) + 1) + 8.0 * vector(paru, (pari) + 2, (parj), (park) + 1) + vector(paru, (pari) - 2, (parj), (park) + 2) + 8.0 * vector(paru, (pari) + 1, (parj), (park) + 2))) / (144.0 * dx[0] * dx[2]))

#define D1CDO4crossed_ji(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (((vector(paru, (pari) - 2, (parj) - 2, (park)) + 8.0 * vector(paru, (pari) - 2, (parj) + 1, (park)) + 64.0 * vector(paru, (pari) - 1, (parj) - 1, (park)) + 8.0 * vector(paru, (pari) - 1, (parj) + 2, (park)) + 8.0 * vector(paru, (pari) + 1, (parj) - 2, (park)) + 64.0 * vector(paru, (pari) + 1, (parj) + 1, (park)) + 8.0 * vector(paru, (pari) + 2, (parj) - 1, (park)) + vector(paru, (pari) + 2, (parj) + 2, (park))) - (8.0 * vector(paru, (pari) - 2, (parj) - 1, (park)) + vector(paru, (pari) - 2, (parj) + 2, (park)) + 8.0 * vector(paru, (pari) - 1, (parj) - 2, (park)) + 64.0 * vector(paru, (pari) - 1, (parj) + 1, (park)) + 64.0 * vector(paru, (pari) + 1, (parj) - 1, (park)) + 8.0 * vector(paru, (pari) + 1, (parj) + 2, (park)) + vector(paru, (pari) + 2, (parj) - 2, (park)) + 8.0 * vector(paru, (pari) + 2, (parj) + 1, (park)))) / (144.0 * dx[1] * dx[0]))

#define D1CDO4crossed_jj(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (((vector(paru, (pari), (parj) - 2, (park)) + 8.0 * vector(paru, (pari), (parj) + 1, (park)) + 64.0 * vector(paru, (pari), (parj) - 1, (park)) + 8.0 * vector(paru, (pari), (parj) + 2, (park)) + 8.0 * vector(paru, (pari), (parj) - 2, (park)) + 64.0 * vector(paru, (pari), (parj) + 1, (park)) + 8.0 * vector(paru, (pari), (parj) - 1, (park)) + vector(paru, (pari), (parj) + 2, (park))) - (8.0 * vector(paru, (pari), (parj) - 1, (park)) + vector(paru, (pari), (parj) + 2, (park)) + 8.0 * vector(paru, (pari), (parj) - 2, (park)) + 64.0 * vector(paru, (pari), (parj) + 1, (park)) + 64.0 * vector(paru, (pari), (parj) - 1, (park)) + 8.0 * vector(paru, (pari), (parj) + 2, (park)) + vector(paru, (pari), (parj) - 2, (park)) + 8.0 * vector(paru, (pari), (parj) + 1, (park)))) / (144.0 * dx[1] * dx[1]))

#define D1CDO4crossed_jk(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (((vector(paru, (pari), (parj) - 2, (park) - 2) + 8.0 * vector(paru, (pari), (parj) + 1, (park) - 2) + 64.0 * vector(paru, (pari), (parj) - 1, (park) - 1) + 8.0 * vector(paru, (pari), (parj) + 2, (park) - 1) + 8.0 * vector(paru, (pari), (parj) - 2, (park) + 1) + 64.0 * vector(paru, (pari), (parj) + 1, (park) + 1) + 8.0 * vector(paru, (pari), (parj) - 1, (park) + 2) + vector(paru, (pari), (parj) + 2, (park) + 2)) - (8.0 * vector(paru, (pari), (parj) - 1, (park) - 2) + vector(paru, (pari), (parj) + 2, (park) - 2) + 8.0 * vector(paru, (pari), (parj) - 2, (park) - 1) + 64.0 * vector(paru, (pari), (parj) + 1, (park) - 1) + 64.0 * vector(paru, (pari), (parj) - 1, (park) + 1) + 8.0 * vector(paru, (pari), (parj) + 2, (park) + 1) + vector(paru, (pari), (parj) - 2, (park) + 2) + 8.0 * vector(paru, (pari), (parj) + 1, (park) + 2))) / (144.0 * dx[1] * dx[2]))

#define D1CDO4crossed_ki(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (((vector(paru, (pari) - 2, (parj), (park) - 2) + 8.0 * vector(paru, (pari) - 2, (parj), (park) + 1) + 64.0 * vector(paru, (pari) - 1, (parj), (park) - 1) + 8.0 * vector(paru, (pari) - 1, (parj), (park) + 2) + 8.0 * vector(paru, (pari) + 1, (parj), (park) - 2) + 64.0 * vector(paru, (pari) + 1, (parj), (park) + 1) + 8.0 * vector(paru, (pari) + 2, (parj), (park) - 1) + vector(paru, (pari) + 2, (parj), (park) + 2)) - (8.0 * vector(paru, (pari) - 2, (parj), (park) - 1) + vector(paru, (pari) - 2, (parj), (park) + 2) + 8.0 * vector(paru, (pari) - 1, (parj), (park) - 2) + 64.0 * vector(paru, (pari) - 1, (parj), (park) + 1) + 64.0 * vector(paru, (pari) + 1, (parj), (park) - 1) + 8.0 * vector(paru, (pari) + 1, (parj), (park) + 2) + vector(paru, (pari) + 2, (parj), (park) - 2) + 8.0 * vector(paru, (pari) + 2, (parj), (park) + 1))) / (144.0 * dx[2] * dx[0]))

#define D1CDO4crossed_kj(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (((vector(paru, (pari), (parj) - 2, (park) - 2) + 8.0 * vector(paru, (pari), (parj) - 2, (park) + 1) + 64.0 * vector(paru, (pari), (parj) - 1, (park) - 1) + 8.0 * vector(paru, (pari), (parj) - 1, (park) + 2) + 8.0 * vector(paru, (pari), (parj) + 1, (park) - 2) + 64.0 * vector(paru, (pari), (parj) + 1, (park) + 1) + 8.0 * vector(paru, (pari), (parj) + 2, (park) - 1) + vector(paru, (pari), (parj) + 2, (park) + 2)) - (8.0 * vector(paru, (pari), (parj) - 2, (park) - 1) + vector(paru, (pari), (parj) - 2, (park) + 2) + 8.0 * vector(paru, (pari), (parj) - 1, (park) - 2) + 64.0 * vector(paru, (pari), (parj) - 1, (park) + 1) + 64.0 * vector(paru, (pari), (parj) + 1, (park) - 1) + 8.0 * vector(paru, (pari), (parj) + 1, (park) + 2) + vector(paru, (pari), (parj) + 2, (park) - 2) + 8.0 * vector(paru, (pari), (parj) + 2, (park) + 1))) / (144.0 * dx[2] * dx[1]))

#define D1CDO4crossed_kk(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (((vector(paru, (pari), (parj), (park) - 2) + 8.0 * vector(paru, (pari), (parj), (park) + 1) + 64.0 * vector(paru, (pari), (parj), (park) - 1) + 8.0 * vector(paru, (pari), (parj), (park) + 2) + 8.0 * vector(paru, (pari), (parj), (park) - 2) + 64.0 * vector(paru, (pari), (parj), (park) + 1) + 8.0 * vector(paru, (pari), (parj), (park) - 1) + vector(paru, (pari), (parj), (park) + 2)) - (8.0 * vector(paru, (pari), (parj), (park) - 1) + vector(paru, (pari), (parj), (park) + 2) + 8.0 * vector(paru, (pari), (parj), (park) - 2) + 64.0 * vector(paru, (pari), (parj), (park) + 1) + 64.0 * vector(paru, (pari), (parj), (park) - 1) + 8.0 * vector(paru, (pari), (parj), (park) + 2) + vector(paru, (pari), (parj), (park) - 2) + 8.0 * vector(paru, (pari), (parj), (park) + 1))) / (144.0 * dx[2] * dx[2]))

#define FVM_i(parfIntPos, parfIntNeg, dx, simPlat_dt, ilast, jlast) (((parfIntPos) - (parfIntNeg)) / dx[0])

#define FVM_j(parfIntPos, parfIntNeg, dx, simPlat_dt, ilast, jlast) (((parfIntPos) - (parfIntNeg)) / dx[1])

#define FVM_k(parfIntPos, parfIntNeg, dx, simPlat_dt, ilast, jlast) (((parfIntPos) - (parfIntNeg)) / dx[2])

#define meshDissipation_i(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) ((vector(paru, (pari) + 3, (parj), (park)) + (-6.0 * vector(paru, (pari) + 2, (parj), (park))) + 15.0 * vector(paru, (pari) + 1, (parj), (park)) + (-20.0 * vector(paru, (pari), (parj), (park))) + 15.0 * vector(paru, (pari) - 1, (parj), (park)) + (-6.0 * vector(paru, (pari) - 2, (parj), (park))) + vector(paru, (pari) - 3, (parj), (park))) / (64.0 * dx[0]))

#define meshDissipation_j(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) ((vector(paru, (pari), (parj) + 3, (park)) + (-6.0 * vector(paru, (pari), (parj) + 2, (park))) + 15.0 * vector(paru, (pari), (parj) + 1, (park)) + (-20.0 * vector(paru, (pari), (parj), (park))) + 15.0 * vector(paru, (pari), (parj) - 1, (park)) + (-6.0 * vector(paru, (pari), (parj) - 2, (park))) + vector(paru, (pari), (parj) - 3, (park))) / (64.0 * dx[1]))

#define meshDissipation_k(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) ((vector(paru, (pari), (parj), (park) + 3) + (-6.0 * vector(paru, (pari), (parj), (park) + 2)) + 15.0 * vector(paru, (pari), (parj), (park) + 1) + (-20.0 * vector(paru, (pari), (parj), (park))) + 15.0 * vector(paru, (pari), (parj), (park) - 1) + (-6.0 * vector(paru, (pari), (parj), (park) - 2)) + vector(paru, (pari), (parj), (park) - 3)) / (64.0 * dx[2]))

#define RK4P1_(parRHS, parQn, dx, simPlat_dt, ilast, jlast) ((parQn) + (simPlat_dt * (parRHS)) / 2.0)

#define eikonal_i(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (MIN(vector(paru, (pari) + 1, (parj), (park)), vector(paru, (pari) - 1, (parj), (park))))

#define eikonal_j(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (MIN(vector(paru, (pari), (parj) + 1, (park)), vector(paru, (pari), (parj) - 1, (park))))

#define eikonal_k(paru, pari, parj, park, dx, simPlat_dt, ilast, jlast) (MIN(vector(paru, (pari), (parj), (park) + 1), vector(paru, (pari), (parj), (park) - 1)))

#define RK4P2_(parRHS, parQn, dx, simPlat_dt, ilast, jlast) ((parQn) + (simPlat_dt * (parRHS)) / 2.0)

#define RK4P3_(parRHS, parQn, dx, simPlat_dt, ilast, jlast) ((parQn) + simPlat_dt * (parRHS))

#define RK4P4_(parRHS, parQn, parQK1, parQK2, parQK3, dx, simPlat_dt, ilast, jlast) ((-(parQn) / 3.0) + (parQK1) / 3.0 + 2.0 * (parQK2) / 3.0 + (parQK3) / 3.0 + (simPlat_dt * (parRHS)) / 6.0)

#define FiSfd_x_mainI(parSfd_x, pargtd_xx, pargtd_xy, pargtd_xz, parBetau_x, parAlpha, parTauT_xx, parTauT_xy, parTauT_xz, parinv_chi, parsdetg, parSfud_xx, dx, simPlat_dt, ilast, jlast) ((-(parBetau_x) * (parSfd_x)) + (parAlpha) * (parSfud_xx) + (-(parAlpha) * (parsdetg) * (parinv_chi) * ((parTauT_xx) * (pargtd_xx) + (parTauT_xy) * (pargtd_xy) + (parTauT_xz) * (pargtd_xz))))

#define FjSfd_x_mainI(parSfd_x, pargtd_xx, pargtd_xy, pargtd_xz, parBetau_y, parAlpha, parTauT_xy, parTauT_yy, parTauT_yz, parinv_chi, parsdetg, parSfud_yx, dx, simPlat_dt, ilast, jlast) ((-(parBetau_y) * (parSfd_x)) + (parAlpha) * (parSfud_yx) + (-(parAlpha) * (parsdetg) * (parinv_chi) * ((parTauT_xy) * (pargtd_xx) + (parTauT_yy) * (pargtd_xy) + (parTauT_yz) * (pargtd_xz))))

#define FkSfd_x_mainI(parSfd_x, pargtd_xx, pargtd_xy, pargtd_xz, parBetau_z, parAlpha, parTauT_xz, parTauT_yz, parTauT_zz, parinv_chi, parsdetg, parSfud_zx, dx, simPlat_dt, ilast, jlast) ((-(parBetau_z) * (parSfd_x)) + (parAlpha) * (parSfud_zx) + (-(parAlpha) * (parsdetg) * (parinv_chi) * ((parTauT_xz) * (pargtd_xx) + (parTauT_yz) * (pargtd_xy) + (parTauT_zz) * (pargtd_xz))))

#define FiSfd_y_mainI(parSfd_y, pargtd_xy, pargtd_yy, pargtd_yz, parBetau_x, parAlpha, parTauT_xx, parTauT_xy, parTauT_xz, parinv_chi, parsdetg, parSfud_xy, dx, simPlat_dt, ilast, jlast) ((-(parBetau_x) * (parSfd_y)) + (parAlpha) * (parSfud_xy) + (-(parAlpha) * (parsdetg) * (parinv_chi) * ((parTauT_xx) * (pargtd_xy) + (parTauT_xy) * (pargtd_yy) + (parTauT_xz) * (pargtd_yz))))

#define FjSfd_y_mainI(parSfd_y, pargtd_xy, pargtd_yy, pargtd_yz, parBetau_y, parAlpha, parTauT_xy, parTauT_yy, parTauT_yz, parinv_chi, parsdetg, parSfud_yy, dx, simPlat_dt, ilast, jlast) ((-(parBetau_y) * (parSfd_y)) + (parAlpha) * (parSfud_yy) + (-(parAlpha) * (parsdetg) * (parinv_chi) * ((parTauT_xy) * (pargtd_xy) + (parTauT_yy) * (pargtd_yy) + (parTauT_yz) * (pargtd_yz))))

#define FkSfd_y_mainI(parSfd_y, pargtd_xy, pargtd_yy, pargtd_yz, parBetau_z, parAlpha, parTauT_xz, parTauT_yz, parTauT_zz, parinv_chi, parsdetg, parSfud_zy, dx, simPlat_dt, ilast, jlast) ((-(parBetau_z) * (parSfd_y)) + (parAlpha) * (parSfud_zy) + (-(parAlpha) * (parsdetg) * (parinv_chi) * ((parTauT_xz) * (pargtd_xy) + (parTauT_yz) * (pargtd_yy) + (parTauT_zz) * (pargtd_yz))))

#define FiSfd_z_mainI(parSfd_z, pargtd_xz, pargtd_yz, pargtd_zz, parBetau_x, parAlpha, parTauT_xx, parTauT_xy, parTauT_xz, parinv_chi, parsdetg, parSfud_xz, dx, simPlat_dt, ilast, jlast) ((-(parBetau_x) * (parSfd_z)) + (parAlpha) * (parSfud_xz) + (-(parAlpha) * (parsdetg) * (parinv_chi) * ((parTauT_xx) * (pargtd_xz) + (parTauT_xy) * (pargtd_yz) + (parTauT_xz) * (pargtd_zz))))

#define FjSfd_z_mainI(parSfd_z, pargtd_xz, pargtd_yz, pargtd_zz, parBetau_y, parAlpha, parTauT_xy, parTauT_yy, parTauT_yz, parinv_chi, parsdetg, parSfud_yz, dx, simPlat_dt, ilast, jlast) ((-(parBetau_y) * (parSfd_z)) + (parAlpha) * (parSfud_yz) + (-(parAlpha) * (parsdetg) * (parinv_chi) * ((parTauT_xy) * (pargtd_xz) + (parTauT_yy) * (pargtd_yz) + (parTauT_yz) * (pargtd_zz))))

#define FkSfd_z_mainI(parSfd_z, pargtd_xz, pargtd_yz, pargtd_zz, parBetau_z, parAlpha, parTauT_xz, parTauT_yz, parTauT_zz, parinv_chi, parsdetg, parSfud_zz, dx, simPlat_dt, ilast, jlast) ((-(parBetau_z) * (parSfd_z)) + (parAlpha) * (parSfud_zz) + (-(parAlpha) * (parsdetg) * (parinv_chi) * ((parTauT_xz) * (pargtd_xz) + (parTauT_yz) * (pargtd_yz) + (parTauT_zz) * (pargtd_zz))))

#define FiBfu_x_mainI(parphif, parAlpha, parchi_max, pargtu_xx, dx, simPlat_dt, ilast, jlast) ((parAlpha) * (parchi_max) * (pargtu_xx) * (parphif))

#define FjBfu_x_mainI(parBfu_x, parBfu_y, parphif, parBetau_x, parBetau_y, parAlpha, parTauM_xy, parchi_max, pargtu_xy, parsdetg, parvfu_x, parvfu_y, dx, simPlat_dt, ilast, jlast) ((parBfu_x) * ((parAlpha) * (parvfu_y) - (parBetau_y)) + (-(parBfu_y) * ((parAlpha) * (parvfu_x) - (parBetau_x))) + (parAlpha) * (parchi_max) * (pargtu_xy) * (parphif) + (parAlpha) * (parsdetg) * (parTauM_xy))

#define FkBfu_x_mainI(parBfu_x, parBfu_z, parphif, parBetau_x, parBetau_z, parAlpha, parTauM_xz, parchi_max, pargtu_xz, parsdetg, parvfu_x, parvfu_z, dx, simPlat_dt, ilast, jlast) ((parBfu_x) * ((parAlpha) * (parvfu_z) - (parBetau_z)) + (-(parBfu_z) * ((parAlpha) * (parvfu_x) - (parBetau_x))) + (parAlpha) * (parchi_max) * (pargtu_xz) * (parphif) + (parAlpha) * (parsdetg) * (parTauM_xz))

#define FiBfu_y_mainI(parBfu_x, parBfu_y, parphif, parBetau_x, parBetau_y, parAlpha, parTauM_xy, parchi_max, pargtu_xy, parsdetg, parvfu_x, parvfu_y, dx, simPlat_dt, ilast, jlast) ((parBfu_y) * ((parAlpha) * (parvfu_x) - (parBetau_x)) + (-(parBfu_x) * ((parAlpha) * (parvfu_y) - (parBetau_y))) + (parAlpha) * (parchi_max) * (pargtu_xy) * (parphif) + (-(parAlpha) * (parsdetg) * (parTauM_xy)))

#define FjBfu_y_mainI(parphif, parAlpha, parchi_max, pargtu_yy, dx, simPlat_dt, ilast, jlast) ((parAlpha) * (parchi_max) * (pargtu_yy) * (parphif))

#define FkBfu_y_mainI(parBfu_y, parBfu_z, parphif, parBetau_y, parBetau_z, parAlpha, parTauM_yz, parchi_max, pargtu_yz, parsdetg, parvfu_y, parvfu_z, dx, simPlat_dt, ilast, jlast) ((parBfu_y) * ((parAlpha) * (parvfu_z) - (parBetau_z)) + (-(parBfu_z) * ((parAlpha) * (parvfu_y) - (parBetau_y))) + (parAlpha) * (parchi_max) * (pargtu_yz) * (parphif) + (parAlpha) * (parsdetg) * (parTauM_yz))

#define FiBfu_z_mainI(parBfu_x, parBfu_z, parphif, parBetau_x, parBetau_z, parAlpha, parTauM_xz, parchi_max, pargtu_xz, parsdetg, parvfu_x, parvfu_z, dx, simPlat_dt, ilast, jlast) ((parBfu_z) * ((parAlpha) * (parvfu_x) - (parBetau_x)) + (-(parBfu_x) * ((parAlpha) * (parvfu_z) - (parBetau_z))) + (parAlpha) * (parchi_max) * (pargtu_xz) * (parphif) + (-(parAlpha) * (parsdetg) * (parTauM_xz)))

#define FjBfu_z_mainI(parBfu_y, parBfu_z, parphif, parBetau_y, parBetau_z, parAlpha, parTauM_yz, parchi_max, pargtu_yz, parsdetg, parvfu_y, parvfu_z, dx, simPlat_dt, ilast, jlast) ((parBfu_z) * ((parAlpha) * (parvfu_y) - (parBetau_y)) + (-(parBfu_y) * ((parAlpha) * (parvfu_z) - (parBetau_z))) + (parAlpha) * (parchi_max) * (pargtu_yz) * (parphif) + (-(parAlpha) * (parsdetg) * (parTauM_yz)))

#define FkBfu_z_mainI(parphif, parAlpha, parchi_max, pargtu_zz, dx, simPlat_dt, ilast, jlast) ((parAlpha) * (parchi_max) * (pargtu_zz) * (parphif))

#define FiDf_mainI(parDf, parBetau_x, parAlpha, parTauN_x, parsdetg, parvfu_x, dx, simPlat_dt, ilast, jlast) (((parAlpha) * (parvfu_x) - (parBetau_x)) * (parDf) + (-(parAlpha) * (parsdetg) * (parTauN_x)))

#define FjDf_mainI(parDf, parBetau_y, parAlpha, parTauN_y, parsdetg, parvfu_y, dx, simPlat_dt, ilast, jlast) (((parAlpha) * (parvfu_y) - (parBetau_y)) * (parDf) + (-(parAlpha) * (parsdetg) * (parTauN_y)))

#define FkDf_mainI(parDf, parBetau_z, parAlpha, parTauN_z, parsdetg, parvfu_z, dx, simPlat_dt, ilast, jlast) (((parAlpha) * (parvfu_z) - (parBetau_z)) * (parDf) + (-(parAlpha) * (parsdetg) * (parTauN_z)))

#define FiDYf_mainI(parDYf, parBetau_x, parAlpha, parTauNe_x, parsdetg, parvfu_x, dx, simPlat_dt, ilast, jlast) (((parAlpha) * (parvfu_x) - (parBetau_x)) * (parDYf) + (-(parAlpha) * (parsdetg) * (parTauNe_x)))

#define FjDYf_mainI(parDYf, parBetau_y, parAlpha, parTauNe_y, parsdetg, parvfu_y, dx, simPlat_dt, ilast, jlast) (((parAlpha) * (parvfu_y) - (parBetau_y)) * (parDYf) + (-(parAlpha) * (parsdetg) * (parTauNe_y)))

#define FkDYf_mainI(parDYf, parBetau_z, parAlpha, parTauNe_z, parsdetg, parvfu_z, dx, simPlat_dt, ilast, jlast) (((parAlpha) * (parvfu_z) - (parBetau_z)) * (parDYf) + (-(parAlpha) * (parsdetg) * (parTauNe_z)))

#define Fitauf_mainI(parDf, partauf, parBetau_x, parAlpha, parTauN_x, parsdetg, parSfu_x, parvfu_x, dx, simPlat_dt, ilast, jlast) ((-(parBetau_x) * (partauf)) + (parAlpha) * ((-(parDf) * (parvfu_x)) + (parSfu_x)) + (parAlpha) * (parsdetg) * (parTauN_x))

#define Fjtauf_mainI(parDf, partauf, parBetau_y, parAlpha, parTauN_y, parsdetg, parSfu_y, parvfu_y, dx, simPlat_dt, ilast, jlast) ((-(parBetau_y) * (partauf)) + (parAlpha) * ((-(parDf) * (parvfu_y)) + (parSfu_y)) + (parAlpha) * (parsdetg) * (parTauN_y))

#define Fktauf_mainI(parDf, partauf, parBetau_z, parAlpha, parTauN_z, parsdetg, parSfu_z, parvfu_z, dx, simPlat_dt, ilast, jlast) ((-(parBetau_z) * (partauf)) + (parAlpha) * ((-(parDf) * (parvfu_z)) + (parSfu_z)) + (parAlpha) * (parsdetg) * (parTauN_z))

#define Fiphif_mainI(parBfu_x, parphif, parBetau_x, parAlpha, parch, dx, simPlat_dt, ilast, jlast) ((parAlpha) * ((parch) * (parch)) * (parBfu_x) + (-(parphif) * (parBetau_x)))

#define Fjphif_mainI(parBfu_y, parphif, parBetau_y, parAlpha, parch, dx, simPlat_dt, ilast, jlast) ((parAlpha) * ((parch) * (parch)) * (parBfu_y) + (-(parphif) * (parBetau_y)))

#define Fkphif_mainI(parBfu_z, parphif, parBetau_z, parAlpha, parch, dx, simPlat_dt, ilast, jlast) ((parAlpha) * ((parch) * (parch)) * (parBfu_z) + (-(parphif) * (parBetau_z)))

#define Unit_MCD(i, j, k) ((((int)i) % 10 == 0 && ((int)j) % 10 == 0 && ((int)k) % 10 == 0) ? 10 : (((int)i) % 9 == 0 && ((int)j) % 9 == 0 && ((int)k) % 9 == 0) ? 9 : (((int)i) % 8 == 0 && ((int)j) % 8 == 0 && ((int)k) % 8 == 0) ? 8 : (((int)i) % 7 == 0 && ((int)j) % 7 == 0 && ((int)k) % 7 == 0) ? 7 : (((int)i) % 6 == 0 && ((int)j) % 6 == 0 && ((int)k) % 6 == 0) ? 6 : (((int)i) % 5 == 0 && ((int)j) % 5 == 0 && ((int)k) % 5 == 0) ? 5 : (((int)i) % 4 == 0 && ((int)j) % 4 == 0 && ((int)k) % 4 == 0) ? 4 : (((int)i) % 3 == 0 && ((int)j) % 3 == 0 && ((int)k) % 3 == 0) ? 3 : (((int)i) % 2 == 0 && ((int)j) % 2 == 0 && ((int)k) % 2 == 0) ? 2 : 1)

inline double Minmod4(double parw, double para, double parb, double parc, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {

	if (equalsEq((parw * para * parb * parc), 0.0)) {
		return 0.0;
	} else {
		return 0.125 * (parw / fabs(parw) + para / fabs(para)) * fabs((parw / fabs(parw) + parb / fabs(parb)) * (parw / fabs(parw) + parc / fabs(parc))) * MIN(fabs(parw), MIN(fabs(para), MIN(fabs(parb), fabs(parc))));
	}

};

inline double Minmod(double para, double parb, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {

	if (equalsEq((para * parb), 0.0)) {
		return 0.0;
	} else {
		return 0.5 * (para / fabs(para) + parb / fabs(parb)) * MIN(fabs(para), fabs(parb));
	}

};

inline double MP5(double param2, double param1, double para, double parap1, double parap2, double paralpha, double parepsilon, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double vnorm, vl, vmp, djm1, dj, djp1, dm4jph, dm4jmh, vul, vav, vmd, vlc, vmin, vmax;

	vnorm = sqrt(param2 * param2 + param1 * param1 + para * para + parap1 * parap1 + parap2 * parap2);
	vl = (2.0 * param2 + (-13.0 * param1) + 47.0 * para + 27.0 * parap1 + (-3.0 * parap2)) / 60.0;
	vmp = para + Minmod(parap1 - para, paralpha * (para - param1), dx, simPlat_dt, ilast, jlast);
	if (lessEq(((vl - para) * (vl - vmp)), (parepsilon * vnorm))) {
		return vl;
	} else {
		djm1 = param2 + (-2.0 * param1) + para;
		dj = param1 + (-2.0 * para) + parap1;
		djp1 = para + (-2.0 * parap1) + parap2;
		dm4jph = Minmod4(4.0 * dj - djp1, 4.0 * djp1 - dj, dj, djp1, dx, simPlat_dt, ilast, jlast);
		dm4jmh = Minmod4(4.0 * dj - djm1, 4.0 * djm1 - dj, dj, djm1, dx, simPlat_dt, ilast, jlast);
		vul = para + paralpha * (para - param1);
		vav = 0.5 * (para + parap1);
		vmd = vav - 0.5 * dm4jph;
		vlc = para + 0.5 * (para - param1) + 4.0 / 3.0 * dm4jmh;
		vmin = MAX(MIN(para, MIN(parap1, vmd)), MIN(para, MIN(vul, vlc)));
		vmax = MIN(MAX(para, MAX(parap1, vmd)), MAX(para, MAX(vul, vlc)));
		return vl + Minmod(vmin - vl, vmax - vl, dx, simPlat_dt, ilast, jlast);
	}

};

inline void EOSCOLD(double parrhof, double parrho_0, double parrho_1, double parrho_2, double para_0, double para_1, double para_2, double para_3, double parK_0, double parK_1, double parK_2, double parK_3, double pargamma_0, double pargamma_1, double pargamma_2, double pargamma_3, double& parselected_gamma, double& parp_cold, double& pareps_cold, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {

	if (parrhof > parrho_2) {
		parselected_gamma = pargamma_3;
		parp_cold = parK_3 * (pow(parrhof, pargamma_3));
		pareps_cold = para_3 + parK_3 / (pargamma_3 - 1.0) * (pow(parrhof, (pargamma_3 - 1.0)));
	} else {
		if (greaterEq(parrhof, parrho_1)) {
			parselected_gamma = pargamma_2;
			parp_cold = parK_2 * (pow(parrhof, pargamma_2));
			pareps_cold = para_2 + parK_2 / (pargamma_2 - 1.0) * (pow(parrhof, (pargamma_2 - 1.0)));
		} else {
			if (greaterEq(parrhof, parrho_0)) {
				parselected_gamma = pargamma_1;
				parp_cold = parK_1 * (pow(parrhof, pargamma_1));
				pareps_cold = para_1 + parK_1 / (pargamma_1 - 1.0) * (pow(parrhof, (pargamma_1 - 1.0)));
			} else {
				parselected_gamma = pargamma_0;
				parp_cold = parK_0 * (pow(parrhof, pargamma_0));
				pareps_cold = para_0 + parK_0 / (pargamma_0 - 1.0) * (pow(parrhof, (pargamma_0 - 1.0)));
			}
		}
	}

};

inline void EOSRecovery(double& parrhof, double& parpf, double& parepsf, double& parvfd_x, double& parvfd_y, double& parvfd_z, double& parsqcs, double& pardpfdrho, double& pardpfdeps, double& pardpfdye, double parrho_0, double parrho_1, double parrho_2, double para_0, double para_1, double para_2, double para_3, double parK_0, double parK_1, double parK_2, double parK_3, double pargamma_0, double pargamma_1, double pargamma_2, double pargamma_3, double parselected_gamma, double pargamma, double parp_cold, double pareps_cold, double& parD, double parD_atm, double& parDf, double parsdetg, double& parh, double& parSfd_x, double& parSfd_y, double& parSfd_z, double& parBu_x, double& parBu_y, double& parBu_z, double parBfu_x, double parBfu_y, double parBfu_z, double& parBfd_x, double& parBfd_y, double& parBfd_z, double parinv_chi, double pargtd_xx, double pargtd_xy, double pargtd_xz, double pargtd_yy, double pargtd_yz, double pargtd_zz, double& parBd_x, double& parBd_y, double& parBd_z, double& partau, double& partauf, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double sqB, W;

	parD = parD_atm;
	parDf = parD * parsdetg;
	parrhof = parD;
	EOSCOLD(parrhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, parselected_gamma, parp_cold, pareps_cold, dx, simPlat_dt, ilast, jlast);
	parpf = parp_cold;
	parepsf = pareps_cold;
	pardpfdeps = 0.0;
	pardpfdye = 0.0;
	pardpfdrho = 0.0;
	parvfd_x = 0.0;
	parvfd_y = 0.0;
	parvfd_z = 0.0;
	parh = parrhof * (1.0 + parepsf) + parpf;
	parsqcs = fabs((parselected_gamma * parp_cold) / parh);
	parSfd_x = 0.0;
	parSfd_y = 0.0;
	parSfd_z = 0.0;
	parBu_x = parBfu_x / parsdetg;
	parBu_y = parBfu_y / parsdetg;
	parBu_z = parBfu_z / parsdetg;
	parBfd_x = parinv_chi * (pargtd_xx * parBfu_x + pargtd_xy * parBfu_y + pargtd_xz * parBfu_z);
	parBfd_y = parinv_chi * (pargtd_xy * parBfu_x + pargtd_yy * parBfu_y + pargtd_yz * parBfu_z);
	parBfd_z = parinv_chi * (pargtd_xz * parBfu_x + pargtd_yz * parBfu_y + pargtd_zz * parBfu_z);
	parBd_x = parBfd_x / parsdetg;
	parBd_y = parBfd_y / parsdetg;
	parBd_z = parBfd_z / parsdetg;
	sqB = parBd_x * parBu_x + parBd_y * parBu_y + parBd_z * parBu_z;
	W = 1.0;
	partau = parh * W * W + (-parpf) + 0.5 * sqB + (-parDf);
	partauf = partau * parsdetg;

};

inline void MAGRecovery(double parsqB_bound, double parrhof, double parpf, double parepsf, double parvfd_x, double parvfd_y, double parvfd_z, double parsqcs, double parrho_0, double parrho_1, double parrho_2, double para_0, double para_1, double para_2, double para_3, double parK_0, double parK_1, double parK_2, double parK_3, double pargamma_0, double pargamma_1, double pargamma_2, double pargamma_3, double parselected_gamma, double parp_cold, double pareps_cold, double parD, double parD_atm, double parDf, double parsdetg, double parh, double parSfd_x, double parSfd_y, double parSfd_z, double& parBu_x, double& parBu_y, double& parBu_z, double& parBfu_x, double& parBfu_y, double& parBfu_z, double& parBfd_x, double& parBfd_y, double& parBfd_z, double parinv_chi, double pargtd_xx, double pargtd_xy, double pargtd_xz, double pargtd_yy, double pargtd_yz, double pargtd_zz, double& parBd_x, double& parBd_y, double& parBd_z, double partau, double partauf, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {

	parBu_x = parBu_x * parsqB_bound;
	parBu_y = parBu_y * parsqB_bound;
	parBu_z = parBu_z * parsqB_bound;
	parBfu_x = parBu_x * parsdetg;
	parBfu_y = parBu_y * parsdetg;
	parBfu_z = parBu_z * parsdetg;
	parBfd_x = parinv_chi * (pargtd_xx * parBfu_x + pargtd_xy * parBfu_y + pargtd_xz * parBfu_z);
	parBfd_y = parinv_chi * (pargtd_xy * parBfu_x + pargtd_yy * parBfu_y + pargtd_yz * parBfu_z);
	parBfd_z = parinv_chi * (pargtd_xz * parBfu_x + pargtd_yz * parBfu_y + pargtd_zz * parBfu_z);
	parBd_x = parBfd_x / parsdetg;
	parBd_y = parBfd_y / parsdetg;
	parBd_z = parBfd_z / parsdetg;

};

inline double dfxEOS(double parrts, double parqvar, double parrvar, double parsvar, double partvar, double pargamma, double parD, double parrho_0, double parrho_1, double parrho_2, double para_0, double para_1, double para_2, double para_3, double parK_0, double parK_1, double parK_2, double parK_3, double pargamma_0, double pargamma_1, double pargamma_2, double pargamma_3, double parselected_gamma, double parp_cold, double pareps_cold, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double dis, ddis;

	dis = fabs(1.0 - (parrts * parrts * parrvar + (2.0 * parrts + parsvar) * (partvar * partvar)) / ((parrts * parrts) * ((parrts + parsvar) * (parrts + parsvar))));
	ddis = 2.0 / ((parrts * parrts * parrts) * ((parrts + parsvar) * (parrts + parsvar) * (parrts + parsvar))) * (parrvar * (parrts * parrts * parrts) + (parsvar * parsvar) * (partvar * partvar) + 3.0 * parsvar * (partvar * partvar) * parrts + 3.0 * (partvar * partvar) * (parrts * parrts));
	return 1.0 + (partvar * partvar) / (parrts * parrts * parrts) + (pargamma - 1.0) / pargamma * (ddis / (2.0 * sqrt(dis)) * (1.0 + pareps_cold) + (-ddis * parrts) + (-dis)) + (-0.5 * ddis * parsvar);

};

inline double dfxEOS_poly(double parrts, double parqvar, double parrvar, double parsvar, double partvar, double pargamma, double parD, double parrho_0, double parrho_1, double parrho_2, double para_0, double para_1, double para_2, double para_3, double parK_0, double parK_1, double parK_2, double parK_3, double pargamma_0, double pargamma_1, double pargamma_2, double pargamma_3, double parselected_gamma, double parp_cold, double pareps_cold, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double rhof, dis, h_cold, ddis;

	dis = fabs(1.0 - (parrts * parrts * parrvar + (2.0 * parrts + parsvar) * partvar * partvar) / (parrts * parrts * ((parrts + parsvar) * (parrts + parsvar))));
	rhof = parD * sqrt(dis);
	EOSCOLD(rhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, parselected_gamma, parp_cold, pareps_cold, dx, simPlat_dt, ilast, jlast);
	h_cold = rhof * (1.0 + pareps_cold) + parp_cold;
	ddis = 2.0 / ((parrts * parrts * parrts) * ((parrts + parsvar) * (parrts + parsvar) * (parrts + parsvar))) * (parrvar * (parrts * parrts * parrts) + (parsvar * parsvar) * (partvar * partvar) + 3.0 * parsvar * (partvar * partvar) * parrts + 3.0 * (partvar * partvar) * (parrts * parrts));
	return 1.0 - (ddis * (parselected_gamma * parp_cold - h_cold)) / (2.0 * dis * dis * parD);

};

inline double fxEOS(double parrts, double parqvar, double parrvar, double parsvar, double partvar, double pargamma, double parD, double parrho_0, double parrho_1, double parrho_2, double para_0, double para_1, double para_2, double para_3, double parK_0, double parK_1, double parK_2, double parK_3, double pargamma_0, double pargamma_1, double pargamma_2, double pargamma_3, double parselected_gamma, double parp_cold, double pareps_cold, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double rhof, dis;

	dis = fabs(1.0 - (parrts * parrts * parrvar + (2.0 * parrts + parsvar) * (partvar * partvar)) / ((parrts * parrts) * ((parrts + parsvar) * (parrts + parsvar))));
	rhof = parD * sqrt(dis);
	EOSCOLD(rhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, parselected_gamma, parp_cold, pareps_cold, dx, simPlat_dt, ilast, jlast);
	return (1.0 - ((pargamma - 1.0) * dis) / pargamma) * parrts + ((pargamma - 1.0) * sqrt(dis)) / pargamma * (1.0 + pareps_cold) - 1.0 + (1.0 - dis / 2.0) * parsvar + (-0.5) * (partvar * partvar) / (parrts * parrts) + (-parqvar) + (-parp_cold / (pargamma * parD));

};

inline double fxEOS_poly(double parrts, double parqvar, double parrvar, double parsvar, double partvar, double pargamma, double parD, double parrho_0, double parrho_1, double parrho_2, double para_0, double para_1, double para_2, double para_3, double parK_0, double parK_1, double parK_2, double parK_3, double pargamma_0, double pargamma_1, double pargamma_2, double pargamma_3, double parselected_gamma, double parp_cold, double pareps_cold, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double rhof, dis, h_cold;

	dis = fabs(1.0 - (parrts * parrts * parrvar + (2.0 * parrts + parsvar) * partvar * partvar) / (parrts * parrts * ((parrts + parsvar) * (parrts + parsvar))));
	rhof = parD * sqrt(dis);
	EOSCOLD(rhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, parselected_gamma, parp_cold, pareps_cold, dx, simPlat_dt, ilast, jlast);
	h_cold = rhof * (1.0 + pareps_cold) + parp_cold;
	return parrts - h_cold / (parD * dis);

};

inline double newtonRaphson(double parf(double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, const double*, const double, const int, const int), double pardf(double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, const double*, const double, const int, const int), double parinitVal, double& parxvarl, double& parxvarh, double parqvar, double parrvar, double parsvar, double partvar, double pargamma, double parrhof, double parD, double parrho_0, double parrho_1, double parrho_2, double para_0, double para_1, double para_2, double para_3, double parK_0, double parK_1, double parK_2, double parK_3, double pargamma_0, double pargamma_1, double pargamma_2, double pargamma_3, double parselected_gamma, double parp_cold, double pareps_cold, double parmax_error, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double rts, dxold, dxvar, fx, dfx, iter;

	rts = parinitVal;
	if ((rts < parxvarl) || (rts > parxvarh) || !equalsEq(rts, rts)) {
		rts = 0.5 * (parxvarl + parxvarh);
	}
	dxold = fabs(parxvarh - parxvarl);
	dxvar = dxold;
	fx = parf(rts, parqvar, parrvar, parsvar, partvar, pargamma, parD, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, parselected_gamma, parp_cold, pareps_cold, dx, simPlat_dt, ilast, jlast);
	dfx = pardf(rts, parqvar, parrvar, parsvar, partvar, pargamma, parD, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, parselected_gamma, parp_cold, pareps_cold, dx, simPlat_dt, ilast, jlast);
	iter = 0.0;
	while (iter < 200.0) {
		if (((((rts - parxvarh) * dfx - fx) * ((rts - parxvarl) * dfx - fx)) > 0.0) || (fabs(2.0 * fx) > fabs(dxold * dfx))) {
			dxold = dxvar;
			dxvar = 0.5 * (parxvarh - parxvarl);
			rts = parxvarl + dxvar;
			if (equalsEq(parxvarl, rts)) {
				iter = 2000.0;
			}
		} else {
			dxold = dxvar;
			dxvar = fx / dfx;
			rts = rts - dxvar;
			if (equalsEq(rts, (rts + dxvar))) {
				iter = 2000.0;
			}
		}
		if (iter < 2000.0) {
			if (fabs(dxvar) < parmax_error) {
				iter = 2000.0;
			} else {
				fx = parf(rts, parqvar, parrvar, parsvar, partvar, pargamma, parD, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, parselected_gamma, parp_cold, pareps_cold, dx, simPlat_dt, ilast, jlast);
				dfx = pardf(rts, parqvar, parrvar, parsvar, partvar, pargamma, parD, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, parselected_gamma, parp_cold, pareps_cold, dx, simPlat_dt, ilast, jlast);
				if (fx < 0.0) {
					parxvarl = rts;
				} else {
					parxvarh = rts;
				}
				iter = iter + 1.0;
			}
		}
	}
	return rts;

};

inline double zbrent(double parfunc(double, double, double, double, double, double, double, double, double, double, double*, double, int, double*, double, int, double*, double, int, double*, int, const double*, const double, const int, const int), double parYE, double parD, double parqvar, double parrvar, double parsvar, double partvar, double partemp_guess, double parenergyShift, double parx1, double parx2, double partol_x, double pareos_epsmin, double* parcoord0_data_Data, double parcoord0_dx_Data, int parcoord0_size_Data, double* parcoord1_data_Data, double parcoord1_dx_Data, int parcoord1_size_Data, double* parcoord2_data_Data, double parcoord2_dx_Data, int parcoord2_size_Data, double* parvars_data_Data, int parvars_size_Data, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double iter, MAX_NEWT_ITER, PREC, a, b, c, temp_guess_a, temp_guess_b, temp_guess_c, fa, fb, fc, keep_iterating, maxerror, d, e, tol1, s, p, q, r, min1, min2;

	//Implementation of Brent’s method to find the root of a function func to accuracy tol_x
	//The root ust be bracketed by x1 and x2
	//Based on W. H. Press et al. 2007, Numerical Recipes in C: The Art of Scientific Computing, 3rd edition,
	//Cambridge University Press, ISBN 978-0-521-88068-8
	iter = 0.0;
	//Termination critera definitions
	MAX_NEWT_ITER = 100.0;
	PREC = 0.000000000000003;
	//Try initial guess
	a = parx1;
	b = parx2;
	c = parx2;
	temp_guess_a = partemp_guess;
	temp_guess_b = partemp_guess;
	temp_guess_c = partemp_guess;
	fa = parfunc(a, parqvar, parrvar, parsvar, partvar, parYE, parD, temp_guess_a, parenergyShift, pareos_epsmin, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data, dx, simPlat_dt, ilast, jlast);
	fb = parfunc(b, parqvar, parrvar, parsvar, partvar, parYE, parD, temp_guess_b, parenergyShift, pareos_epsmin, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data, dx, simPlat_dt, ilast, jlast);
	//Check if the initial guess the correct solution
	if (((fa > 0.0) && (fb > 0.0)) || ((fa < 0.0) && (fb < 0.0))) {
		return b;
	}
	fc = fb;
	temp_guess_c = temp_guess_b;
	keep_iterating = 1.0;
	maxerror = 0.0;
	while (keep_iterating) {
		if (((fb > 0.0) && (fc > 0.0)) || ((fb < 0.0) && (fc < 0.0))) {
			//Reset c, adjust bracket interval d
			c = a;
			fc = fa;
			d = b - a;
			e = d;
			temp_guess_c = temp_guess_a;
		}
		if (fabs(fc) < fabs(fb)) {
			a = b;
			b = c;
			c = a;
			fa = fb;
			fb = fc;
			fc = fa;
			temp_guess_a = temp_guess_b;
			temp_guess_b = temp_guess_c;
			temp_guess_c = temp_guess_a;
		}
		tol1 = 2.0 * PREC * fabs(b) + 0.5 * partol_x;
		maxerror = 0.5 * (c - b);
		//Check convergence
		if (lessEq(fabs(maxerror), tol1) || equalsEq(fb, 0.0)) {
			//Already found
			return b;
		} else {
			if (greaterEq(fabs(e), tol1) && (fabs(fa) > fabs(fb))) {
				//Inverse quadratic interpolation
				s = fb / fa;
				if (equalsEq(a, c)) {
					//Fake quad interpolation
					p = 2.0 * maxerror * s;
					q = 1.0 - s;
				} else {
					//Actual quad interpolation
					q = fa / fc;
					r = fb / fc;
					p = s * (2.0 * maxerror * q * (q - r) - (b - a) * (r - 1.0));
					q = (q - 1.0) * (r - 1.0) * (s - 1.0);
				}
				//Check whether in bounds
				if (p > 0.0) {
					q = -q;
				}
				p = fabs(p);
				min1 = 3.0 * maxerror * q - fabs(tol1 * q);
				min2 = fabs(e * q);
				if ((2.0 * p) < MIN(min1, min2)) {
					//Interpolation successful
					e = d;
					d = p / q;
				} else {
					//Interpolation failed, use bisection
					d = maxerror;
					e = d;
				}
			} else {
				//Bounds decreasing too slowly, use bisection
				d = maxerror;
				e = d;
			}
			//Update a to be last best guess
			a = b;
			fa = fb;
			temp_guess_a = temp_guess_b;
			//Update trial root
			if (fabs(d) > tol1) {
				b = b + d;
			} else {
				if (greaterEq(maxerror, 0.0)) {
					b = b + fabs(tol1);
				} else {
					b = b - fabs(tol1);
				}
			}
			fb = parfunc(b, parqvar, parrvar, parsvar, partvar, parYE, parD, temp_guess_b, parenergyShift, pareos_epsmin, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data, dx, simPlat_dt, ilast, jlast);
		}
		iter = iter + 1.0;
		//Termination criterion
		if (lessEq(fabs(maxerror), tol1) || greaterEq(iter, MAX_NEWT_ITER)) {
			keep_iterating = 0.0;
		}
	}
	if (lessEq(fabs(maxerror), tol1) || equalsEq(fb, 0.0)) {
		return b;
	} else {
		return 0.0;
		TBOX_ERROR("con2prim: error in zbrent");
	}

};

inline double fxTAB(double parrts, double parqvar, double parrvar, double parsvar, double partvar, double parye, double parD, double partemp_guess, double parenergyShift, double pareos_epsmin, double* parcoord0_data_Data, double parcoord0_dx_Data, int parcoord0_size_Data, double* parcoord1_data_Data, double parcoord1_dx_Data, int parcoord1_size_Data, double* parcoord2_data_Data, double parcoord2_dx_Data, int parcoord2_size_Data, double* parvars_data_Data, int parvars_size_Data, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double P, Wminus2, W, rho, eps, lr, lt, lxeps, ltout;

	P = 0.0;
	Wminus2 = 1.0 - (parrts * parrts * parrvar + (2.0 * parrts + parsvar) * partvar * partvar) / (parrts * parrts * (parrts + parsvar) * (parrts + parsvar));
	Wminus2 = fabs(Wminus2);
	W = pow(Wminus2, (-0.5));
	rho = parD / W;
	eps = W - 1.0 + (1.0 - W * W) * parrts / W + W * (parqvar - parsvar + partvar * partvar / (2.0 * parrts * parrts) + parsvar / (2.0 * W * W));
	eps = MAX(eps, pareos_epsmin) + parenergyShift;
	lr = log(rho);
	lt = log(partemp_guess);
	lxeps = log(eps);
	ltout = findCoordFromFile3D_1(3.0, 1.0, lr, lt, parye, lxeps, 1.0, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data);
	readFromFileLinear3D(3.0, 1.0, lr, ltout, parye, P, 2.0, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data);
	P = exp(P);
	eps = eps - parenergyShift;
	return parrts - (1.0 + eps + P / rho) * W;

};

inline void palen1d(double parxvar1, double parxvar2, double parqvar, double parrvar, double parsvar, double partvar, double parBS, double parsqB, double& parTf, double parYef, double parD, double parenergyShift, double& parrhof, double& parepsf, double& parpf, double& parW, double& parsqcs, double& pardpfdrho, double& pardpfdeps, double& pardpfdye, double& parvfd_x, double& parvfd_y, double& parvfd_z, double parBd_x, double parBd_y, double parBd_z, double parSd_x, double parSd_y, double parSd_z, double pareos_epsmin, double* parcoord0_data_Data, double parcoord0_dx_Data, int parcoord0_size_Data, double* parcoord1_data_Data, double parcoord1_dx_Data, int parcoord1_size_Data, double* parcoord2_data_Data, double parcoord2_dx_Data, int parcoord2_size_Data, double* parvars_data_Data, int parvars_size_Data, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double temp_guess, tol, rts, Wminus2, rho, eps, temp, press, lr, lt, lxeps, ltout, cs2, dpdrho, dpdt, dpdye, dedrho, dedt, dedye, dcdrho, dcdt, dcdye, h, dpfdt, defdrho, defdt, defdye, Z;

	//Reference https://arxiv.org/pdf/1403.3680.pdf
	temp_guess = parTf;
	tol = 0.000000005;
	rts = zbrent(fxTAB, parYef, parD, parqvar, parrvar, parsvar, partvar, temp_guess, parenergyShift, parxvar1, parxvar2, tol, pareos_epsmin, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data, dx, simPlat_dt, ilast, jlast);
	Wminus2 = 1.0 - (rts * rts * parrvar + (2.0 * rts + parsvar) * partvar * partvar) / (rts * rts * (rts + parsvar) * (rts + parsvar));
	Wminus2 = fabs(Wminus2);
	parW = pow(Wminus2, (-0.5));
	rho = parD / parW;
	eps = parW - 1.0 + (1.0 - parW * parW) * rts / parW + parW * (parqvar - parsvar + partvar * partvar / (2.0 * rts * rts) + parsvar / (2.0 * parW * parW));
	eps = MAX(eps, pareos_epsmin) + parenergyShift;
	temp = parTf;
	press = 0.0;
	lr = log(rho);
	lt = log(temp);
	lxeps = log(eps);
	//Get temperature for current energy
	ltout = findCoordFromFile3D_1(3.0, 1.0, lr, lt, parYef, lxeps, 1.0, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data);
	temp = exp(ltout);
	cs2 = 0.0;
	dpdrho = 0.0;
	dpdt = 0.0;
	dpdye = 0.0;
	dedrho = 0.0;
	dedt = 0.0;
	dedye = 0.0;
	dcdrho = 0.0;
	dcdt = 0.0;
	dcdye = 0.0;
	readFromFileLinearWithDerivatives3D(3.0, 3.0, lr, ltout, parYef, press, lxeps, cs2, 2.0, 1.0, 0.0, dpdrho, dpdt, dpdye, dedrho, dedt, dedye, dcdrho, dcdt, dcdye, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data);
	press = exp(press);
	lxeps = exp(lxeps);
	parrhof = parD / parW;
	parepsf = lxeps - parenergyShift;
	parTf = temp;
	parpf = press;
	h = (1.0 + parepsf) + parpf / parrhof;
	parsqcs = cs2 / h;
	pardpfdrho = parpf / parrhof * dpdrho;
	dpfdt = parpf / parTf * dpdt;
	pardpfdye = parpf * dpdye;
	defdrho = parepsf / parrhof * dedrho;
	defdt = parepsf / parTf * dedt;
	defdye = parepsf / dedye;
	pardpfdeps = pardpfdrho / defdrho + pardpfdye / defdye + dpfdt / defdt;
	Z = rts * parrhof * parW;
	parvfd_x = (parSd_x + parBS * parBd_x / Z) / (Z + parsqB);
	parvfd_y = (parSd_y + parBS * parBd_y / Z) / (Z + parsqB);
	parvfd_z = (parSd_z + parBS * parBd_z / Z) / (Z + parsqB);

};

inline void calculatePartialDerivatives(double parrho, double partemp, double parye, double& pareps, double& parpress, double& pardedrho, double& pardedt, double& pardpdrho, double& pardpdt, double parenergyShift, double* parcoord0_data_Data, double parcoord0_dx_Data, int parcoord0_size_Data, double* parcoord1_data_Data, double parcoord1_dx_Data, int parcoord1_size_Data, double* parcoord2_data_Data, double parcoord2_dx_Data, int parcoord2_size_Data, double* parvars_data_Data, int parvars_size_Data, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double lr, lt, dpdye, dedye;

	lr = log(parrho);
	lt = log(partemp);
	dpdye = 0.0;
	dedye = 0.0;
	readFromFileLinearWithDerivatives3D(3.0, 2.0, lr, lt, parye, parpress, pareps, 2.0, 1.0, pardpdrho, pardpdt, dpdye, pardedrho, pardedt, dedye, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data);
	parpress = exp(parpress);
	pareps = exp(pareps);
	pardedrho = pareps / parrho * pardedrho;
	pardedt = pareps / partemp * pardedt;
	pardpdrho = parpress / parrho * pardpdrho;
	pardpdt = parpress / partemp * pardpdt;
	pareps = pareps - parenergyShift;

};

inline double con2prim(int pareos_type, double& parrhof, double& parYef, double& parTf, double& parpf, double& parepsf, double& parvfd_x, double& parvfd_y, double& parvfd_z, double& parsqcs, double& pardpfdrho, double& pardpfdeps, double& pardpfdye, double parrho_0, double parrho_1, double parrho_2, double para_0, double para_1, double para_2, double para_3, double parK_0, double parK_1, double parK_2, double parK_3, double pargamma_0, double pargamma_1, double pargamma_2, double pargamma_3, double parD, double parvacuum_D, double& parDf, double& parDYf, double parsdetg, double& parSd_x, double& parSd_y, double& parSd_z, double& parSu_x, double& parSu_y, double& parSu_z, double& parSfd_x, double& parSfd_y, double& parSfd_z, double& partau, double parvacuum_tau, double& partauf, double& parW, double& parh, double parchi, double pargtu_xx, double pargtu_xy, double pargtu_xz, double pargtu_yy, double pargtu_yz, double pargtu_zz, double pargtd_xx, double pargtd_xy, double pargtd_xz, double pargtd_yy, double pargtd_yz, double pargtd_zz, double parBd_x, double parBd_y, double parBd_z, double parBu_x, double parBu_y, double parBu_z, double& parBfu_x, double& parBfu_y, double& parBfu_z, double parmax_error, double pargamma, double parinv_chi, double parthreshold_atm, double parthreshold_sqSmax, double parthreshold_sqBmax, double parenergyShift, double parvacuum_ye_beta, double parvacuum_rho, double parvacuum_rho_reset, double parvacuum_P_reset, double parvacuum_ye_reset, double parvacuum_temp_reset, double parvacuum_tau_reset, double parmin_temp, double parmin_eps, double* parcoord0_data_Data, double parcoord0_dx_Data, int parcoord0_size_Data, double* parcoord1_data_Data, double parcoord1_dx_Data, int parcoord1_size_Data, double* parcoord2_data_Data, double parcoord2_dx_Data, int parcoord2_size_Data, double* parvars_data_Data, int parvars_size_Data, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double xprs, xrho, xtemp, lr, lt, cs2, dpdr, dpde, xenr, P, eps, hh, sqB, sqBmax, sqB_bound, sqS, BS, sqSmax, ss, qvar, rvar, svar, tvar, xvar1, xvar2, dis, fl, fh, iter, rts, xvarl, xvarh, h_cold, Wx, Zx, Tx, count, x_lowlim_W, x_lowlim_Z, x_lowlim_T, NEWT_TOL, MAX_NEWT_ITER, rho, press, x_old_W, x_old_Z, x_old_T, dx_W, dx_Z, dx_T, f_tau, f_S, f_x, error_W, error_Z, error_T, keep_iterating, drhodW, depsdW, dpdeps, depsdP, dPdT, dEdW, dEdZ, dEdT, E, a, b, c, d, e, f, g, k, A, B, C, detJ, maxerror, ye, temp, lxeps, dpdrho, dpdt, dpdye, dedrho, dedt, dedye, dcdrho, dcdt, dcdye, dpfdt, defdrho, defdt, defdye, selected_gamma, p_cold, eps_cold, Bfd_x, Bfd_y, Bfd_z, E_EOS, P_EOS, depsEOSdrho, depsEOSdT, dpEOSdrho, dpEOSdT;

	if (equalsEq(pareos_type, 1.0) && equalsEq(parvacuum_P_reset, 0.0)) {
		TBOX_ERROR("Vacuum parameters are not properly set. Copy them from the first execution to the paremeters file.");
	}
	//eos_type 0: Analytic (Hybrid) EOS
	//eos_type 1: Microphysical(Tabulated) EOS
	if (equalsEq(pareos_type, 1.0) && (parDf < (parthreshold_atm * parvacuum_D * parsdetg))) {
		//Tabulated eos at vacuum
		parrhof = parvacuum_rho_reset;
		parvfd_x = 0.0;
		parvfd_y = 0.0;
		parvfd_z = 0.0;
		parYef = parvacuum_ye_beta;
		parTf = parvacuum_temp_reset;
		xprs = 0.0;
		xrho = parrhof;
		xtemp = parTf;
		lr = log(xrho);
		lt = log(xtemp);
		cs2 = 0.0;
		dpdr = 0.0;
		dpde = 0.0;
		xenr = 0.0;
		readFromFileLinear3D(3.0, 3.0, lr, lt, parYef, xprs, xenr, cs2, 2.0, 1.0, 0.0, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data);
		P = exp(xprs);
		eps = exp(xenr) - parenergyShift;
		parpf = P;
		parepsf = eps;
		parTf = xtemp;
		parh = parrhof * (1.0 + parepsf) + parpf;
		hh = 1.0 + parepsf + parpf / parrhof;
		parsqcs = cs2 / hh;
		if (parsqcs < 0.0) {
			parsqcs = 0.1111111111111111;
		}
		parW = 1.0;
		parDf = parrhof;
		partauf = (parh - parpf) - parDf;
		parSfd_x = 0.0;
		parSfd_y = 0.0;
		parSfd_z = 0.0;
		parBfu_x = 0.0;
		parBfu_y = 0.0;
		parBfu_z = 0.0;
		parDYf = parDf * parYef;
		pardpfdrho = 0.0;
		pardpfdeps = 0.0;
		pardpfdye = 0.0;
		return 0.0;
	}
	if (equalsEq(pareos_type, 0.0) && (parD < (parthreshold_atm * parvacuum_D))) {
		//Hybrid eos at vacuum
		EOSRecovery(parrhof, parpf, parepsf, parvfd_x, parvfd_y, parvfd_z, parsqcs, pardpfdrho, pardpfdeps, pardpfdye, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, pargamma, p_cold, eps_cold, parD, parvacuum_D, parDf, parsdetg, parh, parSfd_x, parSfd_y, parSfd_z, parBu_x, parBu_y, parBu_z, parBfu_x, parBfu_y, parBfu_z, Bfd_x, Bfd_y, Bfd_z, parinv_chi, pargtd_xx, pargtd_xy, pargtd_xz, pargtd_yy, pargtd_yz, pargtd_zz, parBd_x, parBd_y, parBd_z, partau, partauf, dx, simPlat_dt, ilast, jlast);
	} else {
		//Common variables for both kinds of EOS
		sqB = parBd_x * parBu_x + parBd_y * parBu_y + parBd_z * parBu_z;
		sqBmax = fabs(parthreshold_sqBmax * parD);
		if ((sqB > sqBmax) && (sqBmax > 0.0)) {
			sqB_bound = sqrt(sqBmax / sqB);
			MAGRecovery(sqB_bound, parrhof, parpf, parepsf, parvfd_x, parvfd_y, parvfd_z, parsqcs, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, p_cold, eps_cold, parD, parvacuum_D, parDf, parsdetg, parh, parSfd_x, parSfd_y, parSfd_z, parBu_x, parBu_y, parBu_z, parBfu_x, parBfu_y, parBfu_z, Bfd_x, Bfd_y, Bfd_z, parinv_chi, pargtd_xx, pargtd_xy, pargtd_xz, pargtd_yy, pargtd_yz, pargtd_zz, parBd_x, parBd_y, parBd_z, partau, partauf, dx, simPlat_dt, ilast, jlast);
			sqB = parBd_x * parBu_x + parBd_y * parBu_y + parBd_z * parBu_z;
		}
		sqS = parSd_x * parSu_x + parSd_y * parSu_y + parSd_z * parSu_z;
		BS = parSu_x * parBd_x + parSu_y * parBd_y + parSu_z * parBd_z;
		if (equalsEq(pareos_type, 1.0)) {
			if (partauf < (parthreshold_atm * (parvacuum_tau * parsdetg + 0.5 * sqB))) {
				partau = parvacuum_tau_reset + 0.5 * sqB;
				partauf = partau * parsdetg;
			}
		} else {
			if (partau < (parthreshold_atm * (parvacuum_tau + 0.5 * sqB))) {
				partau = parvacuum_tau + 0.5 * sqB;
				partauf = partau * parsdetg;
			}
		}
		if (greaterEq(partau, 0.0)) {
			sqSmax = parthreshold_sqSmax * (parD * 2.0 + partau) * partau;
		} else {
			sqSmax = parthreshold_sqSmax * (parD + partau) * (parD + partau);
		}
		if ((sqS > sqSmax) && (sqSmax > 0.0)) {
			ss = sqrt(sqSmax / sqS);
			parSd_x = parSd_x * ss;
			parSd_y = parSd_y * ss;
			parSd_z = parSd_z * ss;
			parSu_x = parchi * pargtu_xx * parSd_x + parchi * pargtu_xy * parSd_y + parchi * pargtu_xz * parSd_z;
			parSu_y = parchi * pargtu_xy * parSd_x + parchi * pargtu_yy * parSd_y + parchi * pargtu_yz * parSd_z;
			parSu_z = parchi * pargtu_xz * parSd_x + parchi * pargtu_yz * parSd_y + parchi * pargtu_zz * parSd_z;
			sqS = parSd_x * parSu_x + parSd_y * parSu_y + parSd_z * parSu_z;
			BS = parSu_x * parBd_x + parSu_y * parBd_y + parSu_z * parBd_z;
			parSfd_x = parSd_x * parsdetg;
			parSfd_y = parSd_y * parsdetg;
			parSfd_z = parSd_z * parsdetg;
		}
		//Reescaled conserved quanties
		qvar = partau / parD;
		rvar = sqS / (parD * parD);
		svar = sqB / parD;
		tvar = BS / (pow(parD, (3.0 / 2.0)));
		xvar1 = 0.9 * ((1.0 + qvar) - svar);
		xvar2 = 1.1 * (2.0 + 2.0 * qvar);
		if (equalsEq(pareos_type, 0.0)) {
			//Hybrid eos inside the Star
			//Trying to find bounds for the solution for the hybrid EOS
			dis = fabs(1.0 - (xvar1 * xvar1 * rvar + (2.0 * xvar1 + svar) * (tvar * tvar)) / ((xvar1 * xvar1) * ((xvar1 + svar) * (xvar1 + svar))));
			parrhof = parD * sqrt(dis);
			EOSCOLD(parrhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, p_cold, eps_cold, dx, simPlat_dt, ilast, jlast);
			fl = (1.0 - ((pargamma - 1.0) * dis) / pargamma) * xvar1 + ((pargamma - 1.0) * sqrt(dis)) / pargamma * (1.0 + eps_cold) - 1.0 + (1.0 - dis / 2.0) * svar + (-0.5) * (tvar * tvar) / (xvar1 * xvar1) + (-qvar) + (-p_cold / (pargamma * parD));
			dis = fabs(1.0 - (xvar2 * xvar2 * rvar + (2.0 * xvar2 + svar) * (tvar * tvar)) / ((xvar2 * xvar2) * ((xvar2 + svar) * (xvar2 + svar))));
			parrhof = parD * sqrt(dis);
			EOSCOLD(parrhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, p_cold, eps_cold, dx, simPlat_dt, ilast, jlast);
			fh = (1.0 - ((pargamma - 1.0) * dis) / pargamma) * xvar2 + ((pargamma - 1.0) * sqrt(dis)) / pargamma * (1.0 + eps_cold) - 1.0 + (1.0 - dis / 2.0) * svar + (-0.5) * (tvar * tvar) / (xvar2 * xvar2) + (-qvar) + (-p_cold / (pargamma * parD));
			iter = 0.0;
			while (((fl * fh) > 0.0) && (iter < 10.0)) {
				xvar1 = xvar1 * 0.8;
				dis = fabs(1.0 - (xvar1 * xvar1 * rvar + (2.0 * xvar1 + svar) * (tvar * tvar)) / ((xvar1 * xvar1) * ((xvar1 + svar) * (xvar1 + svar))));
				parrhof = parD * sqrt(dis);
				EOSCOLD(parrhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, p_cold, eps_cold, dx, simPlat_dt, ilast, jlast);
				fl = (1.0 - ((pargamma - 1.0) * dis) / pargamma) * xvar1 + ((pargamma - 1.0) * sqrt(dis)) / pargamma * (1.0 + eps_cold) - 1.0 + (1.0 - dis / 2.0) * svar + (-0.5) * (tvar * tvar) / (xvar1 * xvar1) + (-qvar) + (-p_cold / (pargamma * parD));
				xvar2 = xvar2 * 1.2;
				dis = fabs(1.0 - (xvar2 * xvar2 * rvar + (2.0 * xvar2 + svar) * (tvar * tvar)) / ((xvar2 * xvar2) * ((xvar2 + svar) * (xvar2 + svar))));
				parrhof = parD * sqrt(dis);
				EOSCOLD(parrhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, p_cold, eps_cold, dx, simPlat_dt, ilast, jlast);
				fh = (1.0 - ((pargamma - 1.0) * dis) / pargamma) * xvar2 + ((pargamma - 1.0) * sqrt(dis)) / pargamma * (1.0 + eps_cold) - 1.0 + (1.0 - dis / 2.0) * svar + (-0.5) * (tvar * tvar) / (xvar2 * xvar2) + (-qvar) + (-p_cold / (pargamma * parD));
				iter = iter + 1.0;
			}
			if ((fl * fh) < 0.0) {
				//Bounds found, finding solution
				if (equalsEq(fl, 0.0)) {
					rts = xvar1;
				}
				if (equalsEq(fh, 0.0)) {
					rts = xvar2;
				}
				if (fl < 0.0) {
					xvarl = xvar1;
					xvarh = xvar2;
				} else {
					xvarh = xvar1;
					xvarl = xvar2;
				}
				if (!equalsEq(fl, 0.0) && !equalsEq(fh, 0.0)) {
					rts = (parh * parW * parW) / parD;
					rts = newtonRaphson(fxEOS, dfxEOS, rts, xvarl, xvarh, qvar, rvar, svar, tvar, pargamma, parrhof, parD, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, p_cold, eps_cold, parmax_error, dx, simPlat_dt, ilast, jlast);
				}
			} else {
				//Trying to find bounds for the solution for the cold EOS
				qvar = partau / parD;
				rvar = sqS / (parD * parD);
				svar = sqB / parD;
				tvar = BS / (pow(parD, (3.0 / 2.0)));
				xvar1 = (1.0 + qvar) - svar;
				xvar2 = 2.0 + 2.0 * qvar;
				dis = fabs(1.0 - (xvar1 * xvar1 * rvar + (2.0 * xvar1 + svar) * tvar * tvar) / (xvar1 * xvar1 * ((xvar1 + svar) * (xvar1 + svar))));
				parrhof = parD * sqrt(dis);
				EOSCOLD(parrhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, p_cold, eps_cold, dx, simPlat_dt, ilast, jlast);
				h_cold = parrhof * (1.0 + eps_cold) + p_cold;
				fl = xvar1 - h_cold / (parD * dis);
				dis = fabs(1.0 - (xvar2 * xvar2 * rvar + (2.0 * xvar2 + svar) * tvar * tvar) / (xvar2 * xvar2 * ((xvar2 + svar) * (xvar2 + svar))));
				parrhof = parD * sqrt(dis);
				EOSCOLD(parrhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, p_cold, eps_cold, dx, simPlat_dt, ilast, jlast);
				h_cold = parrhof * (1.0 + eps_cold) + p_cold;
				fh = xvar2 - h_cold / (parD * dis);
				iter = 0.0;
				while (((fl * fh) > 0.0) && (iter < 10.0)) {
					xvar1 = xvar1 * 0.5;
					xvar2 = xvar2 * 2.0;
					dis = fabs(1.0 - (xvar1 * xvar1 * rvar + (2.0 * xvar1 + svar) * tvar * tvar) / (xvar1 * xvar1 * ((xvar1 + svar) * (xvar1 + svar))));
					parrhof = parD * sqrt(dis);
					EOSCOLD(parrhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, p_cold, eps_cold, dx, simPlat_dt, ilast, jlast);
					h_cold = parrhof * (1.0 + eps_cold) + p_cold;
					fl = xvar1 - h_cold / (parD * dis);
					dis = fabs(1.0 - (xvar2 * xvar2 * rvar + (2.0 * xvar2 + svar) * tvar * tvar) / (xvar2 * xvar2 * ((xvar2 + svar) * (xvar2 + svar))));
					parrhof = parD * sqrt(dis);
					EOSCOLD(parrhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, p_cold, eps_cold, dx, simPlat_dt, ilast, jlast);
					h_cold = parrhof * (1.0 + eps_cold) + p_cold;
					fh = xvar2 - h_cold / (parD * dis);
					iter = iter + 1.0;
				}
				if ((fl * fh) < 0.0) {
					//Bounds found, finding solution
					if (equalsEq(fl, 0.0)) {
						rts = xvar1;
					}
					if (equalsEq(fh, 0.0)) {
						rts = xvar2;
					}
					if (fl < 0.0) {
						xvarl = xvar1;
						xvarh = xvar2;
					} else {
						xvarl = xvar2;
						xvarh = xvar1;
					}
					if (!equalsEq(fl, 0.0) && !equalsEq(fh, 0.0)) {
						parh = parrhof * (1.0 + parepsf) + parpf;
						rts = (parh * parW * parW) / parD;
						rts = newtonRaphson(fxEOS_poly, dfxEOS_poly, rts, xvarl, xvarh, qvar, rvar, svar, tvar, pargamma, parrhof, parD, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, p_cold, eps_cold, parmax_error, dx, simPlat_dt, ilast, jlast);
					}
				} else {
					//EOS recovery asuming atmosphere
					EOSRecovery(parrhof, parpf, parepsf, parvfd_x, parvfd_y, parvfd_z, parsqcs, pardpfdrho, pardpfdeps, pardpfdye, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, pargamma, p_cold, eps_cold, parD, parvacuum_D, parDf, parsdetg, parh, parSfd_x, parSfd_y, parSfd_z, parBu_x, parBu_y, parBu_z, parBfu_x, parBfu_y, parBfu_z, Bfd_x, Bfd_y, Bfd_z, parinv_chi, pargtd_xx, pargtd_xy, pargtd_xz, pargtd_yy, pargtd_yz, pargtd_zz, parBd_x, parBd_y, parBd_z, partau, partauf, dx, simPlat_dt, ilast, jlast);
					return 0.0;
				}
			}
			dis = fabs(1.0 - (rts * rts * rvar + (2.0 * rts + svar) * (tvar * tvar)) / ((rts * rts) * ((rts + svar) * (rts + svar))));
			parW = 1.0 / sqrt(dis);
			parrhof = parD / parW;
			parh = rts * dis * parD;
			EOSCOLD(parrhof, parrho_0, parrho_1, parrho_2, para_0, para_1, para_2, para_3, parK_0, parK_1, parK_2, parK_3, pargamma_0, pargamma_1, pargamma_2, pargamma_3, selected_gamma, p_cold, eps_cold, dx, simPlat_dt, ilast, jlast);
			parpf = (pargamma - 1.0) / pargamma * (parh + (-parrhof) + (-parrhof * eps_cold)) + p_cold / pargamma;
			parYef = parDYf / parDf;
			parpf = MAX(parpf, p_cold);
			parepsf = eps_cold + (parpf - p_cold) / ((pargamma - 1.0) * parrhof);
			parepsf = MAX(parepsf, eps_cold);
			parTf = parepsf;
			pardpfdrho = (1.0 + selected_gamma - pargamma) * p_cold / parrhof + (pargamma - 1.0) * (parepsf - eps_cold);
			pardpfdeps = (pargamma - 1.0) * parrhof;
			pardpfdye = 0.0;
			parvfd_x = (parSd_x / parD + (tvar * parBd_x) / (rts * sqrt(parD))) / (rts + svar);
			parvfd_y = (parSd_y / parD + (tvar * parBd_y) / (rts * sqrt(parD))) / (rts + svar);
			parvfd_z = (parSd_z / parD + (tvar * parBd_z) / (rts * sqrt(parD))) / (rts + svar);
			if (parepsf > eps_cold) {
				parh = parrhof * (1.0 + parepsf) + parpf;
				parsqcs = fabs((pargamma * parpf + (selected_gamma - pargamma) * p_cold) / parh);
			} else {
				h_cold = parrhof * (1.0 + eps_cold) + p_cold;
				parsqcs = fabs((selected_gamma * p_cold) / h_cold);
			}
		}
		if (equalsEq(pareos_type, 1.0)) {
			//Tabulated eos inside the Star
			//Trying to use a 3D Newton Raphson
			Wx = 0.0;
			Zx = 0.0;
			Tx = parTf;
			count = 0.0;
			x_lowlim_W = 1.0;
			x_lowlim_Z = 0.0;
			x_lowlim_T = parmin_temp;
			NEWT_TOL = 0.000000005;
			MAX_NEWT_ITER = 30.0;
			Wx = MAX(parW, 1.0);
			rho = parD / Wx;
			xrho = rho;
			press = 0.0;
			lr = log(xrho);
			lt = log(parTf);
			readFromFileLinear3D(3.0, 1.0, lr, lt, parYef, press, 2.0, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data);
			P = exp(press);
			parh = 1.0 + parepsf + P / rho;
			Zx = rho * parh * Wx * Wx;
			x_old_W = 0.0;
			x_old_Z = 0.0;
			x_old_T = 0.0;
			dx_W = 0.0;
			dx_Z = 0.0;
			dx_T = 0.0;
			f_tau = 0.0;
			f_S = 0.0;
			f_x = 0.0;
			error_W = 0.0;
			error_Z = 0.0;
			error_T = 0.0;
			if (!equalsEq(Wx, Wx) || !equalsEq(Zx, Zx) || !equalsEq(Tx, Tx)) {
				//3D failed, using 1D Newton Raphson
				palen1d(xvar1, xvar2, qvar, rvar, svar, tvar, BS, sqB, parTf, parYef, parD, parenergyShift, parrhof, parepsf, parpf, parW, parsqcs, pardpfdrho, pardpfdeps, pardpfdye, parvfd_x, parvfd_y, parvfd_z, parBd_x, parBd_y, parBd_z, parSd_x, parSd_y, parSd_z, parmin_eps, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data, dx, simPlat_dt, ilast, jlast);
				return 0.0;
			}
			keep_iterating = 1.0;
			while (keep_iterating) {
				rho = parD / Wx;
				calculatePartialDerivatives(rho, Tx, parYef, E_EOS, P_EOS, depsEOSdrho, depsEOSdT, dpEOSdrho, dpEOSdT, parenergyShift, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data, dx, simPlat_dt, ilast, jlast);
				drhodW = -parD / (Wx * Wx);
				depsdW = (-Zx / (parD * Wx * Wx)) - P_EOS / parD + dpEOSdrho / Wx;
				dpdeps = -parD / Wx;
				depsdP = -Wx / parD;
				dPdT = dpdeps * depsEOSdT - dpEOSdT;
				dEdW = depsdW - depsEOSdrho * drhodW;
				dEdZ = 1.0 / (parD * Wx);
				dEdT = depsdP * dpEOSdT - depsEOSdT;
				E = (-1.0) + Zx / (parD * Wx) - P_EOS * Wx / parD;
				E = MAX(E, parmin_eps);
				a = 2.0 * (((Zx + sqB) * (Zx + sqB) - sqS) - (2.0 * Zx + sqB) * (BS * BS) / (Zx * Zx)) * Wx;
				b = (2.0 * (Zx + sqB) + (2.0 / (Zx * Zx) + 2.0 * sqB / (Zx * Zx * Zx)) * BS * BS) * Wx * Wx - 2.0 * (Zx + sqB);
				c = 0.0;
				d = 2.0 * (partau + (parD - Zx) - sqB + (BS * BS) / (2.0 * Zx * Zx) + P_EOS) * Wx - parD * dpEOSdrho;
				e = ((-1.0) - (BS * BS) / (Zx * Zx * Zx)) * Wx * Wx;
				f = Wx * Wx * dpEOSdT;
				g = dEdW;
				hh = dEdZ;
				k = dEdT;
				f_tau = (((Zx + sqB) * (Zx + sqB) - sqS) - (2.0 * Zx + sqB) * (BS * BS) / (Zx * Zx)) * Wx * Wx - (Zx + sqB) * (Zx + sqB);
				f_S = (partau + (parD - Zx) - sqB + (BS * BS) / (2.0 * Zx * Zx) + P_EOS) * Wx * Wx + 0.5 * sqB;
				f_x = E - E_EOS;
				A = e * k - f * hh;
				B = f * g - d * k;
				C = d * hh - e * g;
				detJ = a * A + b * B + c * C;
				dx_W = ((-A / detJ) * f_tau - (c * hh - b * k) / detJ * f_S) - (b * f - c * e) / detJ * f_x;
				dx_Z = ((-B / detJ) * f_tau - (a * k - c * g) / detJ * f_S) - (c * d - a * f) / detJ * f_x;
				dx_T = ((-C / detJ) * f_tau - (g * b - a * hh) / detJ * f_S) - (a * e - b * d) / detJ * f_x;
				x_old_W = Wx;
				x_old_Z = Zx;
				x_old_T = Tx;
				Wx = fabs(Wx + dx_W - x_lowlim_W) + x_lowlim_W;
				Zx = fabs(Zx + dx_Z - x_lowlim_Z) + x_lowlim_Z;
				Tx = fabs(Tx + dx_T - x_lowlim_T) + x_lowlim_T;
				error_W = fabs((Wx - x_old_W) / Wx);
				error_Z = fabs((Zx - x_old_Z) / Zx);
				error_T = fabs((Tx - x_old_T) / Tx);
				if (!equalsEq(Wx, Wx) || !equalsEq(Zx, Zx) || !equalsEq(Tx, Tx)) {
					//3D failed, using 1D Newton Raphson
					palen1d(xvar1, xvar2, qvar, rvar, svar, tvar, BS, sqB, parTf, parYef, parD, parenergyShift, parrhof, parepsf, parpf, parW, parsqcs, pardpfdrho, pardpfdeps, pardpfdye, parvfd_x, parvfd_y, parvfd_z, parBd_x, parBd_y, parBd_z, parSd_x, parSd_y, parSd_z, parmin_eps, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data, dx, simPlat_dt, ilast, jlast);
					return 0.0;
				}
				maxerror = MAX(error_W, MAX(error_Z, error_T));
				count = count + 1.0;
				if (lessEq(fabs(maxerror), NEWT_TOL) || greaterEq(count, MAX_NEWT_ITER)) {
					keep_iterating = 0.0;
				}
			}
			if (!equalsEq(f_tau, f_tau) || !equalsEq(f_S, f_S) || !equalsEq(f_x, f_x) || (fabs(maxerror) > NEWT_TOL)) {
				//3D failed, using 1D Newton Raphson
				palen1d(xvar1, xvar2, qvar, rvar, svar, tvar, BS, sqB, parTf, parYef, parD, parenergyShift, parrhof, parepsf, parpf, parW, parsqcs, pardpfdrho, pardpfdeps, pardpfdye, parvfd_x, parvfd_y, parvfd_z, parBd_x, parBd_y, parBd_z, parSd_x, parSd_y, parSd_z, parmin_eps, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data, dx, simPlat_dt, ilast, jlast);
				return 0.0;
			}
			rho = parD / Wx;
			ye = parYef;
			temp = Tx;
			lr = log(rho);
			lt = log(temp);
			lxeps = 0.0;
			cs2 = 0.0;
			dpdrho = 0.0;
			dpdt = 0.0;
			dpdye = 0.0;
			dedrho = 0.0;
			dedt = 0.0;
			dedye = 0.0;
			dcdrho = 0.0;
			dcdt = 0.0;
			dcdye = 0.0;
			readFromFileLinearWithDerivatives3D(3.0, 3.0, lr, lt, ye, press, lxeps, cs2, 2.0, 1.0, 0.0, dpdrho, dpdt, dpdye, dedrho, dedt, dedye, dcdrho, dcdt, dcdye, parcoord0_data_Data, parcoord0_dx_Data, parcoord0_size_Data, parcoord1_data_Data, parcoord1_dx_Data, parcoord1_size_Data, parcoord2_data_Data, parcoord2_dx_Data, parcoord2_size_Data, parvars_data_Data, parvars_size_Data);
			press = exp(press);
			eps = exp(lxeps);
			parrhof = parD / Wx;
			parepsf = eps - parenergyShift;
			parTf = temp;
			parYef = ye;
			parpf = press;
			parW = Wx;
			parh = (1.0 + parepsf) + parpf / parrhof;
			parsqcs = cs2 / parh;
			pardpfdrho = parpf / parrhof * dpdrho;
			dpfdt = parpf / parTf * dpdt;
			pardpfdye = parpf * dpdye;
			defdrho = parepsf / parrhof * dedrho;
			defdt = parepsf / parTf * dedt;
			defdye = parepsf / dedye;
			pardpfdeps = pardpfdrho / defdrho + pardpfdye / defdye + dpfdt / defdt;
			parvfd_x = (parSd_x + BS * parBd_x / Zx) / (Zx + sqB);
			parvfd_y = (parSd_y + BS * parBd_y / Zx) / (Zx + sqB);
			parvfd_z = (parSd_z + BS * parBd_z / Zx) / (Zx + sqB);
			return 0.0;
		}
	}
	return 0.0;

};

inline double Smooth(double paredge0, double paredge1, double parinput, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double value;

	value = MAX(0.0, MIN(1.0, (parinput - paredge0) / (paredge1 - paredge0)));
	return value * value * (3.0 - 2.0 * value);

};

inline double eikonal2d(double parvh, double parvv, double parf, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {

	if (lessEq(fabs(parvh - parvv), (dx[0] * parf))) {
		return 0.5 * (parvh + parvv) + 0.5 * sqrt(fabs((parvh + parvv) * (parvh + parvv) - 2.0 * (parvh * parvh + parvv * parvv + (-dx[0] * dx[0] * parf * parf))));
	} else {
		return MIN(parvh, parvv) + dx[0] * parf;
	}

};

inline double eikonal3d(double parvh, double parvv, double parvd, double parf, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double disc;

	disc = (parvh + parvv + parvd) * (parvh + parvv + parvd) - 3.0 * (parvh * parvh + parvv * parvv + parvd * parvd + (-dx[0] * dx[0] * parf * parf));
	if (greaterEq(disc, 0.0)) {
		return (parvh + parvv + parvd) / 3.0 + sqrt(fabs(disc)) / 3.0;
	} else {
		return MIN(eikonal2d(parvh, parvv, parf, dx, simPlat_dt, ilast, jlast), MIN(eikonal2d(parvh, parvd, parf, dx, simPlat_dt, ilast, jlast), eikonal2d(parvv, parvd, parf, dx, simPlat_dt, ilast, jlast)));
	}

};

inline double get_fermi_integral(int parifermi, double paretaIn, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double fermi_integral_analytical, eta;

	fermi_integral_analytical = 0.0;
	eta = MAX(-100.0, MIN(paretaIn, 1000.0));
	if (eta > 0.001) {
		if (equalsEq(parifermi, 0.0)) {
			fermi_integral_analytical = log(1.0 + exp(eta));
		} else {
			if (equalsEq(parifermi, 1.0)) {
				fermi_integral_analytical = ((eta * eta) / 2.0 + 1.6449) / (1.0 + exp(-1.6855 * eta));
			} else {
				if (equalsEq(parifermi, 2.0)) {
					fermi_integral_analytical = ((eta * eta * eta) / 3.0 + 3.2899 * eta) / (1.0 - exp(-1.8246 * eta));
				} else {
					if (equalsEq(parifermi, 3.0)) {
						fermi_integral_analytical = ((eta * eta * eta * eta) / 4.0 + 4.9348 * (eta * eta) + 11.3644) / (1.0 + exp(-1.9039 * eta));
					} else {
						if (equalsEq(parifermi, 4.0)) {
							fermi_integral_analytical = ((eta * eta * eta * eta * eta) / 5.0 + 6.5797 * (eta * eta * eta) + 45.4576 * eta) / (1.0 - exp(-1.9484 * eta));
						} else {
							if (equalsEq(parifermi, 5.0)) {
								fermi_integral_analytical = ((eta * eta * eta * eta * eta * eta) / 6.0 + 8.2247 * (eta * eta * eta * eta) + 113.6439 * (eta * eta) + 236.5323) / (1.0 + exp(-1.9727 * eta));
							}
						}
					}
				}
			}
		}
	} else {
		if (equalsEq(parifermi, 0.0)) {
			fermi_integral_analytical = log(1.0 + exp(eta));
		} else {
			if (equalsEq(parifermi, 1.0)) {
				fermi_integral_analytical = exp(eta) / (1.0 + 0.2159 * exp(0.8857 * eta));
			} else {
				if (equalsEq(parifermi, 2.0)) {
					fermi_integral_analytical = 2.0 * exp(eta) / (1.0 + 0.1092 * exp(0.8908 * eta));
				} else {
					if (equalsEq(parifermi, 3.0)) {
						fermi_integral_analytical = 6.0 * exp(eta) / (1.0 + 0.0559 * exp(0.9069 * eta));
					} else {
						if (equalsEq(parifermi, 4.0)) {
							fermi_integral_analytical = 24.0 * exp(eta) / (1.0 + 0.0287 * exp(0.9257 * eta));
						} else {
							if (equalsEq(parifermi, 5.0)) {
								fermi_integral_analytical = 120.0 * exp(eta) / (1.0 + 0.0147 * exp(0.9431 * eta));
							}
						}
					}
				}
			}
		}
	}
	return MIN(fermi_integral_analytical, 100000000000000000000000.0);

};

inline void leakage(int pardo_leakage, double parvacuum_rho, double parthreshold_leakage_vacuum, double parenergyShift, double& paroptdepthe, double& paroptdeptha, double& paroptdepthx, double& parQnu_e, double& parQnu_a, double& parQnu_x, double& parchie, double& parchia, double& parchix, double& parqnu, double& parrnu, double parrhof, double parepsf, double parYef, double parTf, double parsdetg, double& paroptdeptheh, double& paroptdepthah, double& paroptdepthxh, double& paroptdepthev, double& paroptdepthav, double& paroptdepthxv, double& paroptdepthed, double& paroptdepthad, double& paroptdepthxd, double& parchieh, double& parchiah, double& parchixh, double& parchiev, double& parchiav, double& parchixv, double& parchied, double& parchiad, double& parchixd, double parf_redshift, double* parcoord0_data_DataLeakage, double parcoord0_dx_DataLeakage, int parcoord0_size_DataLeakage, double* parcoord1_data_DataLeakage, double parcoord1_dx_DataLeakage, int parcoord1_size_DataLeakage, double* parcoord2_data_DataLeakage, double parcoord2_dx_DataLeakage, int parcoord2_size_DataLeakage, double* parvars_data_DataLeakage, int parvars_size_DataLeakage, const double* dx, const double simPlat_dt, const int ilast, const int jlast) {
	double lye, lrho, ltrho, ltemp, lttemp, eps, lxeps, xmu_e, xmu_p, xmu_n, xxa, xxh, xxn, xxp, xabar, xzbar, eta_e, eta_p, eta_n, eta_hat, eta_nue, eta_nua, eta_nux, csn_0, csp_0, t1, kappa_const_scat_n, kappa_const_scat_p, kappa_const_abs, local_eta_nux, eta_nue_eq, local_eta_nue, local_eta_nua, xlye, xyn, xyp, xynp, xypn, t2, kappa_scat_ne, kappa_scat_pe, kappa_abs_n, kappa_scat_na, kappa_scat_pa, kappa_abs_p, kappa_scat_nx, kappa_scat_px, kappa_tote, kappa_tota, kappa_totx, ds, scattering_kappa, kappa_tilde_nu_scat_xx, kappa_tilde_nu_scat_xy, kappa_tilde_nu_scat_yx, kappa_tilde_nu_scat_yy, kappa_tilde_nu_scat_zx, kappa_tilde_nu_scat_zy, kappa_tilde_nu_scat_xz, kappa_tilde_nu_scat_yz, kappa_tilde_nu_scat_zz, eta_pn, eta_np, abs_kappa, block_factor, kappa_tilde_nu_abs_xx, kappa_tilde_nu_abs_yx, kappa_tilde_nu_abs_zx, kappa_tilde_nu_abs_xy, kappa_tilde_nu_abs_yy, kappa_tilde_nu_abs_zy, kappa_tilde_nu_abs_xz, kappa_tilde_nu_abs_yz, kappa_tilde_nu_abs_zz, zeta_x, zeta_y, zeta_z, rate_const, Rnu_e_diff, Qnu_e_diff, Rnu_a_diff, Qnu_a_diff, Rnu_x_diff, Qnu_x_diff, beta, Rnu_e_free, Qnu_e_free, Rnu_a_free, Qnu_a_free, Rnu_x_free, Qnu_x_free, fermi_integral3p, fermi_integral3m, fermi_integral4p, fermi_integral4m, exponent_block_factor, block_factor_e, block_factor_a, block_factor_x, enr_m, enr_p, enr_tilde_m, enr_tilde_p, pair_const, R_pair, gamma, gamma_const, R_gamma, mysum, Qnu_e_tmp, Qnu_a_tmp, Qnu_x_tmp, Rnu_e, Rnu_a;

	if (equalsEq(pardo_leakage, 1.0)) {
		if (parrhof > (parthreshold_leakage_vacuum * parvacuum_rho)) {
			paroptdeptheh = MAX(paroptdeptheh, 0.0);
			paroptdepthev = MAX(paroptdepthev, 0.0);
			paroptdepthed = MAX(paroptdepthed, 0.0);
			paroptdepthah = MAX(paroptdepthah, 0.0);
			paroptdepthav = MAX(paroptdepthav, 0.0);
			paroptdepthad = MAX(paroptdepthad, 0.0);
			paroptdepthxh = MAX(paroptdepthxh, 0.0);
			paroptdepthxv = MAX(paroptdepthxv, 0.0);
			paroptdepthxd = MAX(paroptdepthxd, 0.0);
			parchieh = MAX(parchieh, 0.0);
			parchiev = MAX(parchiev, 0.0);
			parchied = MAX(parchied, 0.0);
			parchiah = MAX(parchiah, 0.0);
			parchiav = MAX(parchiav, 0.0);
			parchiad = MAX(parchiad, 0.0);
			parchixh = MAX(parchixh, 0.0);
			parchixv = MAX(parchixv, 0.0);
			parchixd = MAX(parchixd, 0.0);
			lye = parYef;
			lrho = parrhof;
			ltrho = log(lrho);
			ltemp = parTf;
			lttemp = log(ltemp);
			eps = parepsf + parenergyShift;
			lxeps = log(MAX(eps, 1.0));
			xmu_e = 0.0;
			xmu_p = 0.0;
			xmu_n = 0.0;
			xxa = 0.0;
			xxh = 0.0;
			xxn = 0.0;
			xxp = 0.0;
			xabar = 0.0;
			xzbar = 0.0;
			readFromFileLinear3D(3.0, 9.0, ltrho, lttemp, lye, xmu_e, xmu_p, xmu_n, xxa, xxh, xxn, xxp, xabar, xzbar, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, parcoord0_data_DataLeakage, parcoord0_dx_DataLeakage, parcoord0_size_DataLeakage, parcoord1_data_DataLeakage, parcoord1_dx_DataLeakage, parcoord1_size_DataLeakage, parcoord2_data_DataLeakage, parcoord2_dx_DataLeakage, parcoord2_size_DataLeakage, parvars_data_DataLeakage, parvars_size_DataLeakage);
			eta_e = xmu_e / ltemp;
			eta_p = xmu_p / ltemp;
			eta_n = xmu_n / ltemp;
			eta_hat = (eta_n - eta_p) - QNP / ltemp;
			eta_nue = eta_e - eta_n + eta_p;
			eta_nua = -eta_nue;
			eta_nux = 0.0;
			csn_0 = (1.0 + 5.0 * (ALPHA * ALPHA)) / 24.0;
			csp_0 = (4.0 * ((CV - 1.0) * (CV - 1.0)) + 5.0 * (ALPHA * ALPHA)) / 24.0;
			t1 = SIGMA_0 * AVO * lrho * INVRHOGF * ((ltemp / ME_MEV) * (ltemp / ME_MEV));
			kappa_const_scat_n = csn_0 * t1;
			kappa_const_scat_p = csp_0 * t1;
			kappa_const_abs = (1.0 + 3.0 * (ALPHA * ALPHA)) / 4.0 * t1;
			local_eta_nux = 0.0;
			eta_nue_eq = eta_nue;
			local_eta_nue = eta_nue_eq * (1.0 - exp(-paroptdepthe));
			local_eta_nua = (-eta_nue_eq) * (1.0 - exp(-paroptdeptha));
			xlye = lye;
			xyn = (1.0 - xlye) / (1.0 + 2.0 / 3.0 * MAX(eta_n, 0.0));
			xyp = xlye / (1.0 + 2.0 / 3.0 * MAX(eta_p, 0.0));
			t1 = exp(-eta_hat);
			xynp = MAX((2.0 * xlye - 1.0) / (t1 - 1.0), 0.0);
			xypn = MAX(xynp * t1, 0.0);
			t1 = get_fermi_integral(4.0 + 1.0, local_eta_nue, dx, simPlat_dt, ilast, jlast) / MAX(get_fermi_integral(2.0 + 1.0, local_eta_nue, dx, simPlat_dt, ilast, jlast), 0.000000001);
			t2 = 1.0 + exp(eta_e - get_fermi_integral(5.0, local_eta_nue, dx, simPlat_dt, ilast, jlast) / MAX(get_fermi_integral(4.0, local_eta_nue, dx, simPlat_dt, ilast, jlast), 0.000000001));
			kappa_scat_ne = kappa_const_scat_n * xyn * t1;
			kappa_scat_pe = kappa_const_scat_p * xyp * t1;
			kappa_abs_n = kappa_const_abs * xynp * t1 / t2;
			t1 = get_fermi_integral(4.0 + 1.0, local_eta_nua, dx, simPlat_dt, ilast, jlast) / MAX(get_fermi_integral(2.0 + 1.0, local_eta_nua, dx, simPlat_dt, ilast, jlast), 0.000000001);
			t2 = 1.0 + exp((-eta_e) - get_fermi_integral(5.0, local_eta_nua, dx, simPlat_dt, ilast, jlast) / MAX(get_fermi_integral(4.0, local_eta_nua, dx, simPlat_dt, ilast, jlast), 0.000000001));
			kappa_scat_na = kappa_const_scat_n * xyn * t1;
			kappa_scat_pa = kappa_const_scat_p * xyp * t1;
			kappa_abs_p = kappa_const_abs * xypn * t1 / t2;
			t1 = get_fermi_integral(4.0 + 1.0, local_eta_nux, dx, simPlat_dt, ilast, jlast) / MAX(get_fermi_integral(2.0 + 1.0, local_eta_nux, dx, simPlat_dt, ilast, jlast), 0.000000001);
			kappa_scat_nx = kappa_const_scat_n * xyn * t1;
			kappa_scat_px = kappa_const_scat_p * xyp * t1;
			kappa_tote = kappa_scat_pe + kappa_scat_ne + kappa_abs_n;
			kappa_tota = kappa_scat_pa + kappa_scat_na + kappa_abs_p;
			kappa_totx = kappa_scat_px + kappa_scat_nx;
			kappa_tote = MAX(kappa_tote, 0.0);
			kappa_tota = MAX(kappa_tota, 0.0);
			kappa_totx = MAX(kappa_totx, 0.0);
			ds = CODE_LENGTH_UNIT * (pow(parsdetg, (1.0 / 3.0)));
			paroptdepthe = eikonal3d(paroptdeptheh, paroptdepthev, paroptdepthed, kappa_tote * ds, dx, simPlat_dt, ilast, jlast);
			paroptdeptha = eikonal3d(paroptdepthah, paroptdepthav, paroptdepthad, kappa_tota * ds, dx, simPlat_dt, ilast, jlast);
			paroptdepthx = eikonal3d(paroptdepthxh, paroptdepthxv, paroptdepthxd, kappa_totx * ds, dx, simPlat_dt, ilast, jlast);
			scattering_kappa = lrho * INVRHOGF * AVO * 0.25 * SIGMA_0 / (ME_MEV * ME_MEV);
			kappa_tilde_nu_scat_xx = xxn * scattering_kappa;
			kappa_tilde_nu_scat_xy = xxp * scattering_kappa;
			kappa_tilde_nu_scat_yx = kappa_tilde_nu_scat_xx;
			kappa_tilde_nu_scat_yy = kappa_tilde_nu_scat_xy;
			kappa_tilde_nu_scat_zx = kappa_tilde_nu_scat_xx;
			kappa_tilde_nu_scat_zy = kappa_tilde_nu_scat_xy;
			scattering_kappa = lrho * INVRHOGF * AVO * 0.0625 * SIGMA_0 / (ME_MEV * ME_MEV) * xabar * ((1.0 - xzbar / xabar) * (1.0 - xzbar / xabar));
			kappa_tilde_nu_scat_xz = xxh * scattering_kappa;
			kappa_tilde_nu_scat_yz = kappa_tilde_nu_scat_xz;
			kappa_tilde_nu_scat_zz = kappa_tilde_nu_scat_xz;
			eta_pn = AVO * lrho * INVRHOGF * (xxn - xxp) / (exp(eta_hat) - 1.0);
			eta_pn = MAX(eta_pn, 0.0);
			eta_np = AVO * lrho * INVRHOGF * (xxp - xxn) / (exp(-eta_hat) - 1.0);
			eta_np = MAX(eta_np, 0.0);
			if ((lrho * INVRHOGF) < 100000000000.0) {
				eta_pn = AVO * lrho * INVRHOGF * xxp;
				eta_np = AVO * lrho * INVRHOGF * xxn;
			}
			abs_kappa = (1.0 + 3.0 * (ALPHA * ALPHA)) * 0.25 * SIGMA_0 / (ME_MEV * ME_MEV);
			block_factor = 1.0 + exp(eta_e - get_fermi_integral(5.0, eta_nue, dx, simPlat_dt, ilast, jlast) / MAX(get_fermi_integral(4.0, eta_nue, dx, simPlat_dt, ilast, jlast), 0.000000001));
			if (fabs(block_factor) > SMALLNUMBER) {
				kappa_tilde_nu_abs_xx = eta_np * abs_kappa / block_factor;
			} else {
				kappa_tilde_nu_abs_xx = 0.0;
			}
			kappa_tilde_nu_abs_yx = 0.0;
			kappa_tilde_nu_abs_zx = 0.0;
			kappa_tilde_nu_abs_xy = 0.0;
			block_factor = 1.0 + exp((-eta_e) - get_fermi_integral(5.0, eta_nua, dx, simPlat_dt, ilast, jlast) / MAX(get_fermi_integral(4.0, eta_nua, dx, simPlat_dt, ilast, jlast), 0.000000001));
			if (fabs(block_factor) > SMALLNUMBER) {
				kappa_tilde_nu_abs_yy = eta_pn * abs_kappa / block_factor;
			} else {
				kappa_tilde_nu_abs_yy = 0.0;
			}
			kappa_tilde_nu_abs_zy = 0.0;
			kappa_tilde_nu_abs_xz = 0.0;
			kappa_tilde_nu_abs_yz = 0.0;
			kappa_tilde_nu_abs_zz = 0.0;
			zeta_x = kappa_tilde_nu_scat_xx + kappa_tilde_nu_scat_xy + kappa_tilde_nu_scat_xz + kappa_tilde_nu_abs_xx + kappa_tilde_nu_abs_xy + kappa_tilde_nu_abs_xz;
			zeta_y = kappa_tilde_nu_scat_yx + kappa_tilde_nu_scat_yy + kappa_tilde_nu_scat_yz + kappa_tilde_nu_abs_yx + kappa_tilde_nu_abs_yy + kappa_tilde_nu_abs_yz;
			zeta_z = kappa_tilde_nu_scat_zx + kappa_tilde_nu_scat_zy + kappa_tilde_nu_scat_zz + kappa_tilde_nu_abs_zx + kappa_tilde_nu_abs_zy + kappa_tilde_nu_abs_zz;
			parchie = eikonal3d(parchieh, parchiev, parchied, zeta_x * ds, dx, simPlat_dt, ilast, jlast);
			parchia = eikonal3d(parchiah, parchiav, parchiad, zeta_y * ds, dx, simPlat_dt, ilast, jlast);
			parchix = eikonal3d(parchixh, parchixv, parchixd, zeta_z * ds, dx, simPlat_dt, ilast, jlast);
			rate_const = 4.0 * M_PI * CLITE * zeta_x / ((HC_MEVCM * HC_MEVCM * HC_MEVCM) * 6.0 * (parchie * parchie));
			if (fabs(parchie) < SMALLNUMBER) {
				rate_const = 0.0;
			}
			Rnu_e_diff = rate_const * ltemp * get_fermi_integral(0.0, eta_nue, dx, simPlat_dt, ilast, jlast);
			Qnu_e_diff = rate_const * (ltemp * ltemp) * get_fermi_integral(1.0, eta_nue, dx, simPlat_dt, ilast, jlast);
			rate_const = 4.0 * M_PI * CLITE * zeta_y / ((HC_MEVCM * HC_MEVCM * HC_MEVCM) * 6.0 * (parchia * parchia));
			if (fabs(parchia) < SMALLNUMBER) {
				rate_const = 0.0;
			}
			Rnu_a_diff = rate_const * ltemp * get_fermi_integral(0.0, eta_nua, dx, simPlat_dt, ilast, jlast);
			Qnu_a_diff = rate_const * (ltemp * ltemp) * get_fermi_integral(1.0, eta_nua, dx, simPlat_dt, ilast, jlast);
			rate_const = 16.0 * M_PI * CLITE * zeta_z / ((HC_MEVCM * HC_MEVCM * HC_MEVCM) * 6.0 * (parchix * parchix));
			if (fabs(parchix) < SMALLNUMBER) {
				rate_const = 0.0;
			}
			Rnu_x_diff = rate_const * ltemp * get_fermi_integral(0.0, eta_nux, dx, simPlat_dt, ilast, jlast);
			Qnu_x_diff = rate_const * (ltemp * ltemp) * get_fermi_integral(1.0, eta_nux, dx, simPlat_dt, ilast, jlast);
			beta = M_PI * CLITE * (1.0 + 3.0 * (ALPHA * ALPHA)) * SIGMA_0 / ((HC_MEVCM * HC_MEVCM * HC_MEVCM) * (ME_MEV * ME_MEV));
			Rnu_e_free = beta * eta_pn * (ltemp * ltemp * ltemp * ltemp * ltemp) * get_fermi_integral(4.0, eta_e, dx, simPlat_dt, ilast, jlast);
			Qnu_e_free = beta * eta_pn * (ltemp * ltemp * ltemp * ltemp * ltemp * ltemp) * get_fermi_integral(5.0, eta_e, dx, simPlat_dt, ilast, jlast);
			Rnu_a_free = beta * eta_np * (ltemp * ltemp * ltemp * ltemp * ltemp) * get_fermi_integral(4.0, -eta_e, dx, simPlat_dt, ilast, jlast);
			Qnu_a_free = beta * eta_np * (ltemp * ltemp * ltemp * ltemp * ltemp * ltemp) * get_fermi_integral(5.0, -eta_e, dx, simPlat_dt, ilast, jlast);
			Rnu_x_free = 0.0;
			Qnu_x_free = 0.0;
			fermi_integral3p = MAX(get_fermi_integral(3.0, eta_e, dx, simPlat_dt, ilast, jlast), SMALLNUMBER);
			fermi_integral3m = MAX(get_fermi_integral(3.0, -eta_e, dx, simPlat_dt, ilast, jlast), SMALLNUMBER);
			fermi_integral4p = get_fermi_integral(4.0, eta_e, dx, simPlat_dt, ilast, jlast);
			fermi_integral4m = get_fermi_integral(4.0, -eta_e, dx, simPlat_dt, ilast, jlast);
			exponent_block_factor = 0.5 * (fermi_integral4p / fermi_integral3p + fermi_integral4m / fermi_integral3m);
			block_factor_e = 1.0 + exp(eta_nue - exponent_block_factor);
			block_factor_a = 1.0 + exp(eta_nua - exponent_block_factor);
			block_factor_x = 1.0 + exp(eta_nux - exponent_block_factor);
			enr_m = 8.0 * M_PI / (HC_MEVCM * HC_MEVCM * HC_MEVCM) * (ltemp * ltemp * ltemp * ltemp) * fermi_integral3p;
			enr_p = 8.0 * M_PI / (HC_MEVCM * HC_MEVCM * HC_MEVCM) * (ltemp * ltemp * ltemp * ltemp) * fermi_integral3m;
			enr_tilde_m = 8.0 * M_PI / (HC_MEVCM * HC_MEVCM * HC_MEVCM) * (ltemp * ltemp * ltemp * ltemp * ltemp) * fermi_integral4p;
			enr_tilde_p = 8.0 * M_PI / (HC_MEVCM * HC_MEVCM * HC_MEVCM) * (ltemp * ltemp * ltemp * ltemp * ltemp) * fermi_integral4m;
			pair_const = SIGMA_0 * CLITE / (ME_MEV * ME_MEV) * enr_m * enr_p;
			R_pair = pair_const / (36.0 * block_factor_e * block_factor_a) * ((CV - CA) * (CV - CA) + (CV + CA) * (CV + CA));
			Rnu_e_free = Rnu_e_free + R_pair;
			Qnu_e_free = Qnu_e_free + R_pair * 0.5 * (enr_tilde_m * enr_p + enr_m * enr_tilde_p) / (enr_m * enr_p);
			Rnu_a_free = Rnu_a_free + R_pair;
			Qnu_a_free = Qnu_a_free + R_pair * 0.5 * (enr_tilde_m * enr_p + enr_m * enr_tilde_p) / (enr_m * enr_p);
			R_pair = pair_const / (9.0 * (block_factor_x * block_factor_x)) * ((CV - CA) * (CV - CA) + (CV + CA - 2.0) * (CV + CA - 2.0));
			Rnu_x_free = Rnu_x_free + R_pair;
			Qnu_x_free = Qnu_x_free + R_pair * 0.5 * (enr_tilde_m * enr_p + enr_m * enr_tilde_p) / (enr_m * enr_p);
			gamma = GAMMA_0 * sqrt((M_PI * M_PI + 3.0 * (eta_e * eta_e)) / 3.0);
			block_factor_e = 1.0 + exp(eta_nue - (1.0 + 0.5 * (gamma * gamma) / (1.0 + gamma)));
			block_factor_a = 1.0 + exp(eta_nua - (1.0 + 0.5 * (gamma * gamma) / (1.0 + gamma)));
			block_factor_x = 1.0 + exp(eta_nux - (1.0 + 0.5 * (gamma * gamma) / (1.0 + gamma)));
			gamma_const = (M_PI * M_PI * M_PI) * SIGMA_0 * CLITE * (ltemp * ltemp * ltemp * ltemp * ltemp * ltemp * ltemp * ltemp) / ((ME_MEV * ME_MEV) * 3.0 * FSC * (HC_MEVCM * HC_MEVCM * HC_MEVCM * HC_MEVCM * HC_MEVCM * HC_MEVCM)) * (gamma * gamma * gamma * gamma * gamma * gamma) * exp(-gamma) * (1.0 + gamma);
			R_gamma = (CV * CV) * gamma_const / (block_factor_e * block_factor_a);
			Rnu_e_free = Rnu_e_free + R_gamma;
			Qnu_e_free = Qnu_e_free + R_gamma * 0.5 * ltemp * (2.0 + (gamma * gamma) / (1.0 + gamma));
			Rnu_a_free = Rnu_a_free + R_gamma;
			Qnu_a_free = Qnu_a_free + R_gamma * 0.5 * ltemp * (2.0 + (gamma * gamma) / (1.0 + gamma));
			R_gamma = ((CV - 1.0) * (CV - 1.0)) * 4.0 * gamma_const / (block_factor_x * block_factor_x);
			Rnu_x_free = Rnu_x_free + R_gamma;
			Qnu_x_free = Qnu_x_free + R_gamma * 0.5 * ltemp * (2.0 + (gamma * gamma) / (1.0 + gamma));
			mysum = Qnu_e_free + Qnu_e_diff;
			if (fabs(mysum) > SMALLNUMBER) {
				Qnu_e_tmp = Qnu_e_free * Qnu_e_diff / mysum;
			} else {
				Qnu_e_tmp = 0.0;
			}
			mysum = Qnu_a_free + Qnu_a_diff;
			if (fabs(mysum) > SMALLNUMBER) {
				Qnu_a_tmp = Qnu_a_free * Qnu_a_diff / mysum;
			} else {
				Qnu_a_tmp = 0.0;
			}
			mysum = Qnu_x_free + Qnu_x_diff;
			if (fabs(mysum) > SMALLNUMBER) {
				Qnu_x_tmp = Qnu_x_free * Qnu_x_diff / mysum;
			} else {
				Qnu_x_tmp = 0.0;
			}
			mysum = Rnu_e_free + Rnu_e_diff;
			if (fabs(mysum) > SMALLNUMBER) {
				Rnu_e = Rnu_e_free * Rnu_e_diff / mysum;
			} else {
				Rnu_e = 0.0;
			}
			mysum = Rnu_a_free + Rnu_a_diff;
			if (fabs(mysum) > SMALLNUMBER) {
				Rnu_a = Rnu_a_free * Rnu_a_diff / mysum;
			} else {
				Rnu_a = 0.0;
			}
			parQnu_e = Qnu_const * parf_redshift * Qnu_e_tmp * MEV_TO_ERG * (CODE_LENGTH_UNIT * CODE_LENGTH_UNIT * CODE_LENGTH_UNIT);
			parQnu_a = Qnu_const * parf_redshift * Qnu_a_tmp * MEV_TO_ERG * (CODE_LENGTH_UNIT * CODE_LENGTH_UNIT * CODE_LENGTH_UNIT);
			parQnu_x = Qnu_const * parf_redshift * Qnu_x_tmp * MEV_TO_ERG * (CODE_LENGTH_UNIT * CODE_LENGTH_UNIT * CODE_LENGTH_UNIT);
			parqnu = Qnu_e_tmp + Qnu_a_tmp + Qnu_x_tmp;
			parrnu = Rnu_e - Rnu_a;
			parqnu = (-parqnu) * MEV_TO_ERG * RHOGF * EPSGF * CODE_TIME_UNIT;
			parrnu = (-parrnu) * MASSN_CGS * RHOGF * CODE_TIME_UNIT;
		} else {
			parQnu_e = 0.0;
			parQnu_a = 0.0;
			parQnu_x = 0.0;
			parrnu = 0.0;
			parqnu = 0.0;
		}
	}

};

inline void extrapolate_field(double* pard_i_f, int pari, double* pard_j_f, int parj, double* pard_k_f, int park, double* parto_f, double* parFOV, const double* dx, const int ilast, const int jlast) {
	double sign_i_ext, sign_j_ext, sign_k_ext, disp_i_j, disp_i_k, disp_j_i, disp_j_k, disp_k_i, disp_k_j, y1, y2, y3, to_i_f, to_j_f, to_k_f, mod_d, term_i, term_j, term_k;

	sign_i_ext = SIGN(vector(pard_i_f, pari, parj, park));
	sign_j_ext = SIGN(vector(pard_j_f, pari, parj, park));
	sign_k_ext = SIGN(vector(pard_k_f, pari, parj, park));
	disp_i_j = 0.0;
	disp_i_k = 0.0;
	disp_j_i = 0.0;
	disp_j_k = 0.0;
	disp_k_i = 0.0;
	disp_k_j = 0.0;
	if (!equalsEq(vector(pard_i_f, pari, parj, park), 0.0) && equalsEq(vector(parFOV, int(pari - vector(pard_i_f, pari, parj, park)), parj, park), 0.0)) {
		disp_j_i = vector(pard_j_f, pari, parj, park);
		disp_k_i = vector(pard_k_f, pari, parj, park);
	}
	if (!equalsEq(vector(pard_j_f, pari, parj, park), 0.0) && equalsEq(vector(parFOV, pari, int(parj - vector(pard_j_f, pari, parj, park)), park), 0.0)) {
		disp_i_j = vector(pard_i_f, pari, parj, park);
		disp_k_j = vector(pard_k_f, pari, parj, park);
	}
	if (!equalsEq(vector(pard_k_f, pari, parj, park), 0.0) && equalsEq(vector(parFOV, pari, parj, int(park - vector(pard_k_f, pari, parj, park))), 0.0)) {
		disp_i_k = vector(pard_i_f, pari, parj, park);
		disp_j_k = vector(pard_j_f, pari, parj, park);
	}
	if (vector(pard_i_f, pari, parj, park) > 0.0) {
		y1 = vector(parto_f, int(pari - (vector(pard_i_f, pari, parj, park) + 2 * sign_i_ext)), int(parj - disp_j_i), int(park - disp_k_i));
		y2 = vector(parto_f, int(pari - (vector(pard_i_f, pari, parj, park) + sign_i_ext)), int(parj - disp_j_i), int(park - disp_k_i));
		y3 = vector(parto_f, int(pari - vector(pard_i_f, pari, parj, park)), int(parj - disp_j_i), int(park - disp_k_i));
		to_i_f = POLINT_MACRO_QUADRATIC_1(y1, y2, y3, vector(pard_i_f, pari, parj, park), sign_i_ext, dx, simPlat_dt, ilast, jlast);
	}
	if (vector(pard_i_f, pari, parj, park) < 0.0) {
		y3 = vector(parto_f, int(pari - (vector(pard_i_f, pari, parj, park) + 2 * sign_i_ext)), int(parj - disp_j_i), int(park - disp_k_i));
		y2 = vector(parto_f, int(pari - (vector(pard_i_f, pari, parj, park) + sign_i_ext)), int(parj - disp_j_i), int(park - disp_k_i));
		y1 = vector(parto_f, int(pari - vector(pard_i_f, pari, parj, park)), int(parj - disp_j_i), int(park - disp_k_i));
		to_i_f = POLINT_MACRO_QUADRATIC_2(y1, y2, y3, vector(pard_i_f, pari, parj, park), sign_i_ext, dx, simPlat_dt, ilast, jlast);
	}
	if (vector(pard_j_f, pari, parj, park) > 0.0) {
		y1 = vector(parto_f, int(pari - disp_i_j), int(parj - (vector(pard_j_f, pari, parj, park) + 2 * sign_j_ext)), int(park - disp_k_j));
		y2 = vector(parto_f, int(pari - disp_i_j), int(parj - (vector(pard_j_f, pari, parj, park) + sign_j_ext)), int(park - disp_k_j));
		y3 = vector(parto_f, int(pari - disp_i_j), int(parj - vector(pard_j_f, pari, parj, park)), int(park - disp_k_j));
		to_j_f = POLINT_MACRO_QUADRATIC_1(y1, y2, y3, vector(pard_j_f, pari, parj, park), sign_j_ext, dx, simPlat_dt, ilast, jlast);
	}
	if (vector(pard_j_f, pari, parj, park) < 0.0) {
		y3 = vector(parto_f, int(pari - disp_i_j), int(parj - (vector(pard_j_f, pari, parj, park) + 2 * sign_j_ext)), int(park - disp_k_j));
		y2 = vector(parto_f, int(pari - disp_i_j), int(parj - (vector(pard_j_f, pari, parj, park) + sign_j_ext)), int(park - disp_k_j));
		y1 = vector(parto_f, int(pari - disp_i_j), int(parj - vector(pard_j_f, pari, parj, park)), int(park - disp_k_j));
		to_j_f = POLINT_MACRO_QUADRATIC_2(y1, y2, y3, vector(pard_j_f, pari, parj, park), sign_j_ext, dx, simPlat_dt, ilast, jlast);
	}
	if (vector(pard_k_f, pari, parj, park) > 0.0) {
		y1 = vector(parto_f, int(pari - disp_i_k), int(parj - disp_j_k), int(park - (vector(pard_k_f, pari, parj, park) + 2 * sign_k_ext)));
		y2 = vector(parto_f, int(pari - disp_i_k), int(parj - disp_j_k), int(park - (vector(pard_k_f, pari, parj, park) + sign_k_ext)));
		y3 = vector(parto_f, int(pari - disp_i_k), int(parj - disp_j_k), int(park - vector(pard_k_f, pari, parj, park)));
		to_k_f = POLINT_MACRO_QUADRATIC_1(y1, y2, y3, vector(pard_k_f, pari, parj, park), sign_k_ext, dx, simPlat_dt, ilast, jlast);
	}
	if (vector(pard_k_f, pari, parj, park) < 0.0) {
		y3 = vector(parto_f, int(pari - disp_i_k), int(parj - disp_j_k), int(park - (vector(pard_k_f, pari, parj, park) + 2 * sign_k_ext)));
		y2 = vector(parto_f, int(pari - disp_i_k), int(parj - disp_j_k), int(park - (vector(pard_k_f, pari, parj, park) + sign_k_ext)));
		y1 = vector(parto_f, int(pari - disp_i_k), int(parj - disp_j_k), int(park - vector(pard_k_f, pari, parj, park)));
		to_k_f = POLINT_MACRO_QUADRATIC_2(y1, y2, y3, vector(pard_k_f, pari, parj, park), sign_k_ext, dx, simPlat_dt, ilast, jlast);
	}
	mod_d = sqrt(vector(pard_i_f, pari, parj, park) * vector(pard_i_f, pari, parj, park) + vector(pard_j_f, pari, parj, park) * vector(pard_j_f, pari, parj, park) + vector(pard_k_f, pari, parj, park) * vector(pard_k_f, pari, parj, park));
	term_i = 0.0;
	if (!equalsEq(vector(pard_i_f, pari, parj, park), 0.0)) {
		term_i = (to_i_f - vector(parto_f, int(pari - vector(pard_i_f, pari, parj, park)), int(parj - disp_j_i), int(park - disp_k_i))) / dx[0];
	}
	term_j = 0.0;
	if (!equalsEq(vector(pard_j_f, pari, parj, park), 0.0)) {
		term_j = (to_j_f - vector(parto_f, int(pari - disp_i_j), int(parj - vector(pard_j_f, pari, parj, park)), int(park - disp_k_j))) / dx[1];
	}
	term_k = 0.0;
	if (!equalsEq(vector(pard_k_f, pari, parj, park), 0.0)) {
		term_k = (to_k_f - vector(parto_f, int(pari - disp_i_k), int(parj - disp_j_k), int(park - vector(pard_k_f, pari, parj, park)))) / dx[2];
	}
	vector(parto_f, pari, parj, park) = vector(parto_f, int(pari - vector(pard_i_f, pari, parj, park)), int(parj - vector(pard_j_f, pari, parj, park)), int(park - vector(pard_k_f, pari, parj, park))) + (term_i + term_j + term_k) * sqrt((vector(pard_i_f, pari, parj, park) * dx[0]) * (vector(pard_i_f, pari, parj, park) * dx[0]) + (vector(pard_j_f, pari, parj, park) * dx[1]) * (vector(pard_j_f, pari, parj, park) * dx[1]) + (vector(pard_k_f, pari, parj, park) * dx[2]) * (vector(pard_k_f, pari, parj, park) * dx[2])) / mod_d;

};

void readTable(std::string inputFile, double **coord0, int &coord0Size, double &coord0Dx, double **coord1, int &coord1Size, double &coord1Dx, double **coord2, int &coord2Size, double &coord2Dx, double** data, int& nVars);
void readTable(std::string inputFile, double **coord0, int &coord0Size, double &coord0Dx, double **coord1, int &coord1Size, double &coord1Dx, double **coord2, int &coord2Size, double &coord2Dx, double** data, int& nVars);




#endif
